# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric Library for Python
"""

from __future__ import unicode_literals
from __future__ import absolute_import

from contextlib import contextmanager

from fabric.api import sudo, run, prefix, cd, settings, env
from fabric.contrib.files import exists, append

from rattail.fablib import mkdir
from rattail.fablib import apt


def install_pip():
    """
    Install/upgrade the Pip installer for Python.
    """
    with settings(warn_only=True):
        result = sudo('which pip')
    if result.failed:
        apt.install('python-setuptools')
        sudo('easy_install pip')
        sudo('apt-get --assume-yes purge python-setuptools')
    apt.install('libffi-dev')
    pip('setuptools')
    pip('pip', 'wheel', 'ndg-httpsclient')


def pip(*packages):
    """
    Install one or more packages via ``pip install``.
    """
    packages = ["'{0}'".format(p) for p in packages]
    sudo('pip install --upgrade {0}'.format(' '.join(packages)))


def install_virtualenvwrapper(workon_home='/srv/envs', user='root'):
    """
    Install the `virtualenvwrapper`_ system, with the given ``workon`` home,
    owned by the given user.
    """
    mkdir(workon_home, owner=user)
    pip('virtualenvwrapper')

    configure_virtualenvwrapper('root', workon_home)
    if user != 'root':
        configure_virtualenvwrapper(user, workon_home)
    configure_virtualenvwrapper(env.user, workon_home)


def configure_virtualenvwrapper(user, workon_home='/srv/envs'):
    """
    Configure virtualenvwrapper for the given user account.
    """
    home = sudo('echo $HOME', user=user)
    home = home.rstrip('/')

    def update(script):
        script = '{0}/{1}'.format(home, script)
        if not exists(script):
            sudo('touch {0}'.format(script))
            sudo('chown {0}:{0} {1}'.format(user, script))
        append(script, 'export WORKON_HOME={0}'.format(workon_home), use_sudo=True)
        append(script, 'source /usr/local/bin/virtualenvwrapper.sh', use_sudo=True)

    update('.profile')
    update('.bashrc')


def mkvirtualenv(name, user=None, upgrade_pip=True):
    """
    Make a new Python virtual environment.
    """
    sudo('mkvirtualenv {0}'.format(name))
    if upgrade_pip:
        with workon(name):
            pip('pip', 'setuptools', 'wheel', 'ndg-httpsclient')
    if user:
        with cdvirtualenv(name):
            mkdir('app/log', owner='{0}:{0}'.format(user))


@contextmanager
def workon(name):
    """
    Context manager to prefix your command(s) with the ``workon`` command.
    """
    with prefix('workon {0}'.format(name)):
        yield


@contextmanager
def cdvirtualenv(name, subdirs=[], workon_home='/srv/envs'):
    """
    Context manager to prefix your command(s) with the ``cdvirtualenv`` command.
    """
    if isinstance(subdirs, basestring):
        subdirs = [subdirs]
    path = '{0}/{1}'.format(workon_home, name)
    if subdirs:
        path = '{0}/{1}'.format(path, '/'.join(subdirs))
    with workon(name):
        with cd(path):
            yield



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor catalog parser for KeHE Distributors
"""

from __future__ import unicode_literals

import os
import re
import datetime
from decimal import Decimal

import xlrd
from sqlalchemy.orm import joinedload

from rattail.vendors.catalogs import CatalogParser
from rattail.db.batch.vendorcatalog import VendorCatalogRow
from rattail.gpc import GPC
from rattail.db import model
from rattail.db.cache import cache_model


class KeheCatalogParser(CatalogParser):
    """
    Vendor catalog parser for KeHE Distributors.
    """
    key = 'rattail.contrib.kehe'
    display = "KeHE Distributors"
    vendor_key = 'kehe'

    def parse_effective_date(self, path):
        """
        This parser does not expect to find an effective date within the
        catalog file itself, but it makes a feeble attempt at gleaning it from
        the filename.  The example used to create this code was:

        .. code-block:: none

           DC22 CNP Catalog - February 2015.xls
        """
        basename = os.path.basename(path)
        match = re.search(r' (\w{3} \d{4})\.xls$', basename)
        if match:
            return datetime.datetime.strptime(match.group(1), '%b %Y').date()

    def parse_rows(self, path):
        book = xlrd.open_workbook(path)
        sheet = book.sheet_by_index(0)
        products = cache_model(self.session, model.Product, key='upc',
                               query_options=[joinedload(model.Product.costs),
                                              joinedload(model.Product.cost)])

        for r in range(1, sheet.nrows): # Skip first header row.

            row = VendorCatalogRow()
            upc = sheet.cell_value(r, 5) or None
            row.upc = GPC(int(upc)) if upc else None
            row.brand_name = sheet.cell_value(r, 1)
            row.description = sheet.cell_value(r, 2)
            row.size = sheet.cell_value(r, 3)
            row.vendor_code = '{0:07d}'.format(int(sheet.cell_value(r, 0)))

            # KeHE catalogs sort of have two columns to represent case size.
            # Sometimes the "Case Pack" column is what we want; other times
            # it's the "Un of Sls" column.  There appears to be no way to
            # really know which should be used, so we must query the database
            # to determine which case size is already in effect for the
            # product.  If that doesn't work, then we're back to guessing; we
            # use the "Case Pack" column in that case.
            case_pack = int(sheet.cell_value(r, 6))
            unit_of_sales = int(sheet.cell_value(r, 7))
            if row.upc:
                product = products.get(row.upc)
                if product:

                    # First we try to find an existing cost record for KeHE.  If
                    # one exists, we will honor its case size.
                    cost = product.cost_for_vendor(self.vendor)
                    if cost:
                        if cost.case_size == case_pack:
                            row.case_size = case_pack
                        elif cost.case_size == unit_of_sales:
                            row.case_size = unit_of_sales

                    # If there was no KeHE cost record, then we'll honor the
                    # "preferred" cost record case size, if it has one.
                    if row.case_size is None and product.cost:
                        if product.cost.case_size == case_pack:
                            row.case_size = case_pack
                        elif product.cost.case_size == unit_of_sales:
                            row.case_size = unit_of_sales

            # If all else fails, use the "Case Pack" column.
            if row.case_size is None:
                row.case_size = case_pack

            row.unit_cost = Decimal('{0:0.4f}'.format(sheet.cell_value(r, 8)))
            row.case_cost = row.unit_cost * row.case_size
            yield row



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import warnings
from unittest import TestCase

from sqlalchemy import create_engine
from sqlalchemy.exc import SAWarning

from rattail.db import model
from rattail.db import Session


warnings.filterwarnings(
    'ignore',
    r"^Dialect sqlite\+pysqlite does \*not\* support Decimal objects natively\, "
    "and SQLAlchemy must convert from floating point - rounding errors and other "
    "issues may occur\. Please consider storing Decimal numbers as strings or "
    "integers on this platform for lossless storage\.$",
    SAWarning, r'^sqlalchemy\..*$')


class DataTestCase(TestCase):

    engine_url = os.environ.get('RATTAIL_TEST_ENGINE_URL', 'sqlite://')

    def setUp(self):
        self.engine = create_engine(self.engine_url)
        model.Base.metadata.create_all(bind=self.engine)
        Session.configure(bind=self.engine)
        self.session = Session()
        self.extra_setup()

    def extra_setup(self):
        """
        Derivative classes may define this as necessary, to avoid having to
        override the :meth:`setUp()` method.
        """

    def tearDown(self):
        self.session.close()
        Session.configure(bind=None)
        model.Base.metadata.drop_all(bind=self.engine)

        # # TODO: This doesn't seem to be necessary, hopefully that's good?
        # for table in list(model.Base.metadata.sorted_tables):
        #     if table.name.startswith('batch.'):
        #         model.Base.metadata.remove(table)

        # TODO: Unfortunately this *does* seem to be necessary...
        model.Batch._rowclasses.clear()



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core Importer Stuff
"""

from __future__ import unicode_literals, absolute_import

import logging

from sqlalchemy.orm.exc import NoResultFound

from rattail.db import model
from rattail.core import Object
from rattail.util import load_object
from rattail.db.cache import cache_model


log = logging.getLogger(__name__)


def make_importer(config, session, spec):
    """
    Create an importer instance according to the given spec.  For now..see the
    source code for more details.
    """
    importer = None
    if '.' not in spec and ':' not in spec:
        from rattail.db.importing import models
        if hasattr(models, spec):
            importer = getattr(models, spec)
        elif hasattr(models, '{0}Importer'.format(spec)):
            importer = getattr(models, '{0}Importer'.format(spec))
    else:
        importer = load_object(spec)
    if importer:
        return importer(config, session)


class Importer(Object):
    """
    Base class for model importers.
    """
    supported_fields = []
    normalizer_class = None
    cached_data = None

    complex_fields = []
    """
    Sequence of field names which are considered complex and therefore require
    custom logic provided by the derived class, etc.
    """

    def __init__(self, config, session, **kwargs):
        self.config = config
        self.session = session
        super(Importer, self).__init__(**kwargs)

    @property
    def model_module(self):
        """
        Reference to a module which contains all available / necessary data
        models.  By default this is ``rattail.db.model``.
        """
        return model

    @property
    def model_class(self):
        return getattr(model, self.__class__.__name__[:-8])

    @property
    def model_name(self):
        return self.model_class.__name__

    @property
    def simple_fields(self):
        return self.supported_fields

    def import_data(self, records, fields, key, count=None, max_updates=None, progress=None):
        """
        Import some data.
        """
        if count is None:
            count = len(records)
        if count == 0:
            return [], []

        self.fields = fields
        self.key = key
        if isinstance(key, basestring):
            self.key = (key,)
        self.progress = progress
        self.normalizer = self.normalizer_class() if self.normalizer_class else None
        self.setup()
        self.cache_data(progress)

        # Normalize to remove duplicate source records.  This is more for the
        # sake of sanity since duplicates typically lead to a ping-pong effect
        # where an update-less import is impossible.
        normalized = {}
        for src_data in records:
            key = self.get_key(src_data)
            if key in normalized:
                log.warning("duplicate records from {0}:{1} for key: {2}".format(
                        self.provider.__class__.__module__, self.provider.__class__.__name__, repr(key)))
            normalized[key] = src_data
        records = []
        for key in sorted(normalized):
            records.append(normalized[key])

        prog = None
        if progress:
            prog = progress("Importing {0} data".format(self.model_name), count)

        created = []
        updated = []

        affected = 0
        keys_seen = set()
        for i, src_data in enumerate(records, 1):
            key = self.get_key(src_data)
            if key in keys_seen:
                log.warning("duplicate records from {0}:{1} for key: {2}".format(
                        self.provider.__class__.__module__, self.provider.__class__.__name__, repr(key)))
            else:
                keys_seen.add(key)
            self.normalize_record(src_data)

            dirty = False
            inst_data = self.get_instance_data(src_data)
            if inst_data:
                if self.data_differs(inst_data, src_data):
                    instance = self.get_instance(src_data)
                    self.update_instance(instance, src_data, inst_data)
                    updated.append(instance)
                    dirty = True
            else:
                instance = self.new_instance(src_data)
                assert instance, "Failed to create new model instance for data: {0}".format(repr(src_data))
                self.update_instance(instance, src_data)
                self.session.add(instance)
                self.session.flush()
                log.debug("created new {} {}: {}".format(self.model_name, key, instance))
                created.append(instance)
                dirty = True

            if dirty:
                self.session.flush()
                affected += 1
                if max_updates and affected >= max_updates:
                    log.warning("max of {0} updates has been reached; bailing early".format(max_updates))
                    break

            if prog:
                prog.update(i)
        if prog:
            prog.destroy()

        return created, updated

    def setup(self):
        """
        Perform any setup necessary, e.g. cache lookups for existing data.
        """

    def cache_query_options(self):
        """
        Return a list of options to apply to the cache query, if needed.
        """

    def cache_model(self, model_class, key, **kwargs):
        """
        Convenience method for caching a model.
        """
        kwargs.setdefault('progress', self.progress)
        return cache_model(self.session, model_class, key=key, **kwargs)

    def get_cache_key(self, instance, normalized):
        """
        Get the primary model cache key for a given instance/data object.
        """
        return tuple(normalized['data'].get(k) for k in self.key)

    def cache_data(self, progress):
        """
        Cache all existing model instances as normalized data.
        """
        self.cached_data = self.cache_model(self.model_class, self.get_cache_key,
                                       query_options=self.cache_query_options(),
                                       normalizer=self.normalize_cache)

    def normalize_cache(self, instance):
        """
        Normalizer for cache data.  This adds the instance to the cache in
        addition to its normalized data.  This is so that if lots of updates
        are required, we don't we have to constantly fetch them.
        """
        return {'instance': instance, 'data': self.normalize_instance(instance)}

    def data_differs(self, inst_data, src_data):
        """
        Compare source record data to instance data to determine if there is a
        net change.
        """
        for field in self.fields:
            if src_data[field] != inst_data[field]:
                log.debug("field {0} differed for instance data: {1}, source data: {2}".format(
                        field, repr(inst_data), repr(src_data)))
                return True
        return False

    def string_or_null(self, data, *fields):
        """
        For each field specified, ensure the data value is a non-empty string,
        or ``None``.
        """
        for field in fields:
            if field in data:
                value = data[field]
                value = value.strip() if value else None
                data[field] = value or None

    def int_or_null(self, data, *fields):
        """
        For each field specified, ensure the data value is a non-zero integer,
        or ``None``.
        """
        for field in fields:
            if field in data:
                value = data[field]
                value = int(value) if value else None
                data[field] = value or None

    def prioritize_2(self, data, field):
        """
        Prioritize the data values for the pair of fields implied by the given
        fieldname.  I.e., if only one non-empty value is present, make sure
        it's in the first slot.
        """
        field2 = '{0}_2'.format(field)
        if field in data and field2 in data:
            if data[field2] and not data[field]:
                data[field], data[field2] = data[field2], None

    def normalize_record(self, data):
        """
        Normalize the source data record, if necessary.
        """

    def get_key(self, data):
        """
        Return the key value for the given source data record.
        """
        return tuple(data.get(k) for k in self.key)

    def get_instance(self, data):
        """
        Fetch an instance from our database which corresponds to the source
        data, if possible; otherwise return ``None``.
        """
        key = self.get_key(data)
        if not key:
            log.warning("source {0} has no {1}: {2}".format(
                    self.model_name, self.key, repr(data)))
            return None

        if self.cached_data is not None:
            data = self.cached_data.get(key)
            return data['instance'] if data else None

        q = self.session.query(self.model_class)
        for i, k in enumerate(self.key):
            q = q.filter(getattr(self.model_class, k) == key[i])
        try:
            instance = q.one()
        except NoResultFound:
            return None
        else:
            return instance

    def get_instance_data(self, data):
        """
        Return a normalized data record for the model instance corresponding to
        the source data record, or ``None``.
        """
        key = self.get_key(data)
        if not key:
            log.warning("source {0} has no {1}: {2}".format(
                    self.model_name, self.key, repr(data)))
            return None
        if self.cached_data is not None:
            data = self.cached_data.get(key)
            return data['data'] if data else None
        instance = self.get_instance(data)
        if instance:
            return self.normalize_instance(instance)

    def normalize_instance(self, instance):
        """
        Normalize a model instance.
        """
        if self.normalizer:
            return self.normalizer.normalize(instance)

        data = {}
        for field in self.simple_fields:
            if field in self.fields:
                data[field] = getattr(instance, field)
        return data

    def new_instance(self, data):
        """
        Return a new model instance to correspond to the source data record.
        """
        kwargs = {}
        key = self.get_key(data)
        for i, k in enumerate(self.key):
            if k in self.simple_fields:
                kwargs[k] = key[i]
        return self.model_class(**kwargs)

    def update_instance(self, instance, data, inst_data=None):
        """
        Update the given model instance with the given data.
        """
        for field in self.simple_fields:
            if field in data:
                if not inst_data or inst_data[field] != data[field]:
                    setattr(instance, field, data[field])


class RecordRenderer(object):
    """
    Record renderer for email notifications sent from data import jobs.
    """

    def __init__(self, config):
        self.config = config

    def __call__(self, record):
        return self.render(record)

    def render(self, record):
        """
        Render the given record.  Default is to attempt.
        """
        key = record.__class__.__name__.lower()
        renderer = getattr(self, 'render_{0}'.format(key), None)
        if renderer:
            return renderer(record)

        label = self.get_label(record)
        url = self.get_url(record)
        if url:
            return '<a href="{0}">{1}</a>'.format(url, label)
        return label

    def get_label(self, record):
        key = record.__class__.__name__.lower()
        label = getattr(self, 'label_{0}'.format(key), self.label)
        return label(record)

    def label(self, record):
        return unicode(record)

    def get_url(self, record):
        """
        Fetch / generate a URL for the given data record.  You should *not*
        override this method, but do :meth:`url()` instead.
        """
        key = record.__class__.__name__.lower()
        url = getattr(self, 'url_{0}'.format(key), self.url)
        return url(record)

    def url(self, record):
        """
        Fetch / generate a URL for the given data record.
        """
        url = self.config.get('tailbone', 'url')
        if url:
            url = url.rstrip('/')
            name = '{0}s'.format(record.__class__.__name__.lower())
            if name == 'persons': # FIXME, obviously this is a hack
                name = 'people'
            url = '{0}/{1}/{{uuid}}'.format(url, name)
            return url.format(uuid=record.uuid)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Development Commands
"""

from __future__ import unicode_literals, absolute_import

import os
import re
import sys

from mako.template import Template

from rattail import commands
from rattail.files import resource_path


def main(*args):
    """
    Entry point for ``rattail-dev`` command.
    """
    Command().run(*args or sys.argv[1:])


class Command(commands.Command):
    """
    Rattail - development commands
    """
    name = 'rattail-dev'
    description = __doc__.strip()


class NewBatch(commands.Subcommand):
    """
    Generate some code for a new batch type.
    """
    name = 'new-batch'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--output-dir', '-O', default='.',
                            help="Path to which generated files should be written.")
        parser.add_argument('model',
                            help="Name of primary model for new batch, e.g. "
                            "'VendorCatalog' or 'PrintLabels'.")

    def run(self, args):
        model_title = re.sub(r'([a-z])([A-Z])', r'\g<1> \g<2>', args.model)
        table_name = model_title.lower().replace(' ', '_')
        context = {
            'model_name': args.model,
            'model_title': model_title,
            'table_name': table_name,
        }
        template_dir = resource_path('rattail:data/new-batch')
        for name in ('__init__', 'model', 'handler', 'webview'):
            template_path = os.path.join(template_dir, '{}.py'.format(name))
            template = Template(filename=template_path)
            output_path = os.path.join(args.output_dir, '{}.py'.format(name))
            with open(output_path, 'wt') as output_file:
                output_file.write(template.render(**context))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase
from mock import patch, DEFAULT, Mock, MagicMock, call

from . import DataTestCase

from rattail.db import changes
from rattail.db import model
from sqlalchemy.orm import RelationshipProperty


class TestChangeRecorder(TestCase):

    def test_init(self):
        recorder = changes.ChangeRecorder()
        self.assertTrue(recorder.ignore_role_changes)
        recorder = changes.ChangeRecorder(False)
        self.assertFalse(recorder.ignore_role_changes)

    # def test_record_change(self):
    #     session = Mock()
    #     recorder = changes.ChangeRecorder()
    #     recorder.ensure_uuid = Mock()

    #     # don't record changes for changes
    #     self.assertFalse(recorder.record_change(session, model.Change()))

    #     # don't record changes for batch data
    #     self.assertFalse(recorder.record_change(session, model.Batch()))
    #     self.assertFalse(recorder.record_change(session, model.BatchColumn()))
    #     self.assertFalse(recorder.record_change(session, model.BatchRow()))

    #     # don't record changes for objects with no uuid attribute
    #     self.assertFalse(recorder.record_change(session, object()))

    #     # don't record changes for role data if so configured
    #     recorder.ignore_role_changes = True
    #     self.assertFalse(recorder.record_change(session, model.Role()))
    #     self.assertFalse(recorder.record_change(session, model.UserRole()))

    #     # none of the above should have involved a call to `ensure_uuid()`
    #     self.assertFalse(recorder.ensure_uuid.called)

    #     # make sure role data is *not* ignored if so configured
    #     recorder.ignore_role_changes = False
    #     self.assertTrue(recorder.record_change(session, model.Role()))
    #     self.assertTrue(recorder.record_change(session, model.UserRole()))

    #     # so far no *new* changes have been created
    #     self.assertFalse(session.add.called)

    #     # mock up session to force new change creation
    #     session.query.return_value = session
    #     session.get.return_value = None
    #     self.assertTrue(recorder.record_change(session, model.Product()))

    @patch.multiple('rattail.db.changes', get_uuid=DEFAULT, object_mapper=DEFAULT)
    def test_ensure_uuid(self, get_uuid, object_mapper):
        recorder = changes.ChangeRecorder()
        uuid_column = Mock()
        object_mapper.return_value.columns.__getitem__.return_value = uuid_column

        # uuid already present
        product = model.Product(uuid='some_uuid')
        recorder.ensure_uuid(product)
        self.assertEqual(product.uuid, 'some_uuid')
        self.assertFalse(get_uuid.called)

        # no uuid yet, auto-generate
        uuid_column.foreign_keys = False
        get_uuid.return_value = 'another_uuid'
        product = model.Product()
        self.assertTrue(product.uuid is None)
        recorder.ensure_uuid(product)
        get_uuid.assert_called_once_with()
        self.assertEqual(product.uuid, 'another_uuid')

        # some heavy mocking for following tests
        uuid_column.foreign_keys = True
        remote_side = MagicMock(key='uuid')
        prop = MagicMock(__class__=RelationshipProperty, key='foreign_thing')
        prop.remote_side.__len__.return_value = 1
        prop.remote_side.__iter__.return_value = [remote_side]
        object_mapper.return_value.iterate_properties.__iter__.return_value = [prop]
        
        # uuid fetched from existing foreign key object
        get_uuid.reset_mock()
        instance = Mock(uuid=None, foreign_thing=Mock(uuid='secondary_uuid'))
        recorder.ensure_uuid(instance)
        self.assertFalse(get_uuid.called)
        self.assertEqual(instance.uuid, 'secondary_uuid')

        # foreign key object doesn't exist; uuid generated as fallback
        get_uuid.return_value = 'fallback_uuid'
        instance = Mock(uuid=None, foreign_thing=None)
        recorder.ensure_uuid(instance)
        get_uuid.assert_called_once_with()
        self.assertEqual(instance.uuid, 'fallback_uuid')


class TestFunctionalChanges(DataTestCase):

    def setUp(self):
        super(TestFunctionalChanges, self).setUp()
        changes.record_changes(self.session)

    def test_add(self):
        product = model.Product()
        self.session.add(product)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.instance_uuid, product.uuid)
        self.assertFalse(change.deleted)

    def test_change(self):
        product = model.Product()
        self.session.add(product)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        product.description = 'Acme Bricks'
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.instance_uuid, product.uuid)
        self.assertFalse(change.deleted)

    def test_delete(self):
        product = model.Product()
        self.session.add(product)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        self.session.delete(product)

        self.assertEqual(self.session.query(model.Change).count(), 1)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.instance_uuid, product.uuid)
        self.assertTrue(change.deleted)

    def test_orphan_change(self):
        department = model.Department()
        subdepartment = model.Subdepartment()
        department.subdepartments.append(subdepartment)
        self.session.add(department)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 2)
        change = self.session.query(model.Change).filter_by(class_name='Department').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='Subdepartment').one()
        self.assertFalse(change.deleted)

        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        # Creating an orphaned Subdepartment, which should be recorded as a
        # *change* due to the cascade rules in effect.
        department.subdepartments.remove(subdepartment)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 2)
        change = self.session.query(model.Change).filter_by(class_name='Department').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='Subdepartment').one()
        self.assertFalse(change.deleted)
        self.assertEqual(self.session.query(model.Subdepartment).count(), 1)
    
    def test_orphan_delete(self):
        customer = model.Customer()
        group = model.CustomerGroup()
        customer.groups.append(group)
        self.session.add(customer)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 3)
        change = self.session.query(model.Change).filter_by(class_name='Customer').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='CustomerGroup').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='CustomerGroupAssignment').one()
        self.assertFalse(change.deleted)

        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        # Creating an orphaned CustomerGroupAssociation, which should be
        # recorded as a *deletion* due to the cascade rules in effect.  Note
        # that the CustomerGroup is not technically an orphan and in fact is
        # not even changed.
        customer.groups.remove(group)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 2)
        change = self.session.query(model.Change).filter_by(class_name='Customer').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='CustomerGroupAssignment').one()
        self.assertTrue(change.deleted)
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase

from rattail import config


class TestParseList(TestCase):

    def test_none(self):
        value = config.parse_list(None)
        self.assertEqual(len(value), 0)

    def test_single_value(self):
        value = config.parse_list(u'foo')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'foo')

    def test_single_value_padded_by_spaces(self):
        value = config.parse_list(u'   foo   ')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'foo')

    def test_slash_is_not_a_separator(self):
        value = config.parse_list(u'/dev/null')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'/dev/null')

    def test_multiple_values_separated_by_whitespace(self):
        value = config.parse_list(u'foo bar baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_commas(self):
        value = config.parse_list(u'foo,bar,baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_whitespace_and_commas(self):
        value = config.parse_list(u'  foo,   bar   baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_whitespace_and_commas_with_some_quoting(self):
        value = config.parse_list(u"""
        foo
        "C:\\some path\\with spaces\\and, a comma",
        baz
""")
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'C:\\some path\\with spaces\\and, a comma')
        self.assertEqual(value[2], u'baz')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Utilities
"""

from __future__ import unicode_literals, absolute_import

import re
import pprint
import logging

from sqlalchemy import orm

# TODO: Deprecate/remove these imports.
from alembic.util import obfuscate_url_pw
from rattail.db.config import engine_from_config, get_engines, get_default_engine, configure_session


log = logging.getLogger(__name__)


def maxlen(attr):
    """
    Return the maximum length for the given attribute.
    """
    if len(attr.property.columns) == 1:
        type_ = attr.property.columns[0].type
        return getattr(type_, 'length', None)


def make_topo_sortkey(model, metadata=None):
    """
    Returns a function suitable for use as a ``key`` kwarg to a standard Python
    sorting call.  This key function will expect a single class mapper and
    return a sequence number associated with that model.  The sequence is
    determined by SQLAlchemy's topological table sorting.
    """
    if metadata is None:
        metadata = model.Base.metadata

    tables = {}
    for i, table in enumerate(metadata.sorted_tables, 1):
        tables[table.name] = i

    log.debug("topo sortkeys for '{}' will be:\n{}".format(model.__name__, pprint.pformat(
        [(i, name) for name, i in sorted(tables.iteritems(), key=lambda t: t[1])])))

    def sortkey(name):
        mapper = orm.class_mapper(getattr(model, name))
        return tuple(tables[t.name] for t in mapper.tables)

    return sortkey


def normalize_full_name(first_name, last_name):
    """
    Normalize the given first and last name to a "full" name value.  The
    fallback return value is an empty string.
    """
    first_name = (first_name or '').strip()
    last_name = (last_name or '').strip()
    if first_name and last_name:
        return "{} {}".format(first_name, last_name)
    if first_name:
        return first_name
    if last_name:
        return last_name
    return ''


def normalize_phone_number(number):
    """
    Normalize a phone number to digits only.
    """
    if number is not None:
        return re.sub(r'\D', '', number)


def format_phone_number(number):
    """
    Returns a phone number in ``(XXX) XXX-XXXX`` format if possible; otherwise
    returns the argument unaltered.
    """
    original, number = number, normalize_phone_number(number)
    if number and len(number) == 10:
        return '({}) {}-{}'.format(number[:3], number[3:6], number[6:])
    return original



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Basic Batch Handler
"""

from __future__ import unicode_literals

import os
import shutil
import datetime

from rattail.db import Session
from rattail.db import model
from rattail.filemon import Action
from rattail.util import load_object


class BatchHandler(object):
    """
    Base class for all batch handlers.  This isn't really useful by itself but
    it is expected that other batches will derive from it.
    """
    show_progress = False

    def __init__(self, config):
        self.config = config

    @property
    def batch_model_class(self):
        """
        Reference to the data model class of the batch type for which this
        handler is responsible.
        """
        raise NotImplementedError("Must set the 'batch_model_class' attribute "
                                  "for class '{0}'".format(self.__class__.__name__))

    def make_batch(self, session, **kwargs):
        """
        Create a new batch instance and return it.  All keyword arguments are
        passed to the batch model constructor.  Note that some keyword
        arguments may be required, depending on the type of batch.
        """
        return self.batch_model_class(**kwargs)

    def get_execute_title(self, batch):
        """
        Get a human-friendly string describing the execution step for a batch.
        Most handlers should probably override this to provide something more
        useful than the default, which is just "Execute this batch".
        """
        return "Execute this batch"

    def refresh_data(self, session, batch, progress=None):
        """
        Refresh all data for the batch.
        """
        del batch.data_rows[:]

    def make_rows(self, session, batch, data, progress=None):
        """
        Create batch rows from the given data set.
        """
        prog = None
        if progress:
            prog = progress("Refreshing data for batch", len(data))

        cancel = False
        for i, row in enumerate(data, 1):
            row.sequence = i
            batch.data_rows.append(row)
            if self.cognize_row(session, row) is not False:
                if i % 250 == 0:    # seems to help progres UI
                    session.flush()
            else:
                batch.data_rows.remove(row)
            if prog and not prog.update(i):
                cancel = True
                break

        if prog:
            prog.destroy()
        return not cancel

    def cognize_row(self, session, row):
        """
        This method should further populate the row's fields, using database
        lookups and business rules etc. as needed.  Each handler class must
        define this method.

        :param session: SQLAlchemy database session object.

        :param row: A batch row instance, whose fields reflect the initial
           source data only, e.g. that which was parsed from a file.

        :returns: Typically this method needn't return anything.  However if it
           returns ``False`` then the row will *not* be added to the batch.
        """
        raise NotImplementedError

    def executable(self, batch):
        """
        This method should return a boolean indicating whether or not execution
        should be allowed for the batch, given its current condition.  The
        default simply returns ``True`` but you may override as needed.

        Note that this (currently) only affects the enabled/disabled state of
        the Execute button within the Tailbone batch view.
        """
        return True

    def execute(self, batch, progress=None):
        raise NotImplementedError


class FileBatchHandler(BatchHandler):
    """
    Base class for all file-based batch handlers.  Adds some conveniences for
    managing data file storage.

    .. note::
       Current implementation only supports one data file per batch.
    """

    @property
    def root_datadir(self):
        """
        The absolute path of the root folder in which data for this particular
        type of batch is stored.  The structure of this path is as follows:

        .. code-block:: none

           /{root_batch_data_dir}/{batch_type_key}

        * ``{root_batch_data_dir}`` - Value of the 'batch.files' option in the
          [rattail] section of config file.
        * ``{batch_type_key}`` - Unique key for the type of batch it is.

        .. note::
           While it is likely that the data folder returned by this method
           already exists, this method does not guarantee it.
        """
        return os.path.join(self.config.require('rattail', 'batch.files'),
                            self.batch_model_class.batch_key)

    def datadir(self, batch):
        """
        Returns the absolute path of the folder in which the batch's source
        data file(s) resides.  Note that the batch must already have been
        persisted to the database.  The structure of the path returned is as
        follows:

        .. code-block:: none

           /{root_datadir}/{uuid[:2]}/{uuid[2:]}

        * ``{root_datadir}`` - Value returned by :meth:`root_datadir()`.
        * ``{uuid[:2]}`` - First two characters of batch UUID.
        * ``{uuid[2:]}`` - All batch UUID characters *after* the first two.

        .. note::
           While it is likely that the data folder returned by this method
           already exists, this method does not guarantee any such thing.  It
           is typically assumed that the path will have been created by a
           previous call to :meth:`make_batch()` however.
        """
        return os.path.join(self.root_datadir, batch.uuid[:2], batch.uuid[2:])

    def data_path(self, batch):
        """
        Returns the full path to the batch's one and only data file.  As with
        :meth:`datadir()`, this method does not guarantee the existence of the
        file.
        """
        return os.path.join(self.datadir(batch), batch.filename)

    def make_batch(self, session, path, **kwargs):
        """
        Create a new batch as per usual, plus save a copy of the data file (at
        ``path``) to the configured batch storage folder.
        """
        kwargs.setdefault('filename', 'tmp')
        batch = self.batch_model_class(**kwargs)
        session.add(batch)
        session.flush()
        self.set_data_file(batch, path)
        return batch

    def set_data_file(self, batch, path):
        """
        Assign the data file found at ``path`` to the batch.  This overwrites
        the batch's :attr:`filename` attribute and places a copy of the data
        file in the batch's data folder.
        """
        batch.filename = os.path.basename(path)
        datadir = self.datadir(batch)
        os.makedirs(datadir)
        shutil.copyfile(path, os.path.join(datadir, batch.filename))


class MakeFileBatch(Action):
    """
    Filemon action for making new file-based batches.
    """

    def __call__(self, path, handler='', user=''):
        """
        Make a batch from the given file path.

        :param path: Path to the source data file for the batch.

        :param handler: Spec string for the batch handler class, e.g.
           ``'rattail.db.batch.vendorcatalog.handler:VendorCatalogHandler'``.

        :param user: Username of the user which is to be credited with creating
           and cognizing the batch.
        """
        handler = load_object(handler)(self.config)
        session = Session()
        user = session.query(model.User).filter_by(username=user).one()
        batch = handler.make_batch(session, path, created_by=user)
        handler.refresh_data(session, batch)
        batch.cognized = datetime.datetime.utcnow()
        batch.cognized_by = user
        session.commit()
        session.close()



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Configuration
"""

from __future__ import unicode_literals

import logging

from rattail.util import load_object
from rattail.exceptions import SQLAlchemyNotInstalled


log = logging.getLogger(__name__)


def engine_from_config(configuration, prefix='sqlalchemy.', **kwargs):
    """
    Custom version of the identically-named SQLAlchemy function.

    The purpose of the customization is to allow the ``poolclass`` parameter to
    be specified within the configuration.  If this option is present in the
    configuration dictionary, it will be coerced to the Python class it
    references and passed as a keyword argument to the "upstream" function.

    An example of the configuration leveraging this feature:

    .. code-block:: ini

       [rattail.db]
       sqlalchemy.url = sqlite:///tmp/rattail.sqlite
       sqlalchemy.poolclass = sqlalchemy.pool:NullPool
    """
    try:
        from sqlalchemy import engine_from_config
    except ImportError as error:
        raise SQLAlchemyNotInstalled(error)

    config = dict(configuration)

    # Convert 'poolclass' arg to actual class.
    key = '{}poolclass'.format(prefix)
    if key in config:
        kwargs.setdefault('poolclass', load_object(config[key]))
        del config[key]

    # Store 'record_changes' flag as private attr on engine.
    record_changes = False
    key = '{}record_changes'.format(prefix)
    if key in config:
        value = config.pop(key)
        # TODO: this logic was copied from `RattailConfig.getbool()`
        record_changes = value.lower() in ('true', 'yes', 'on', '1')

    engine = engine_from_config(config, prefix, **kwargs)
    if record_changes:
        engine._rattail_record_changes = True
    return engine


def get_engines(config, section='rattail.db'):
    """
    Fetch all database engines defined within a given config section.

    :returns: A dictionary of SQLAlchemy engine instances, with keys matching
       those found in config.
    """
    keys = config.get(section, 'keys', usedb=False)
    if keys:
        keys = keys.split(',')
    else:
        keys = ['default']

    engines = {}
    cfg = config.get_dict(section)
    for key in keys:
        key = key.strip()
        try:
            engines[key] = engine_from_config(cfg, prefix='{0}.'.format(key))
        except KeyError:
            if key == 'default':
                try:
                    engines[key] = engine_from_config(cfg, prefix='sqlalchemy.')
                except KeyError:
                    pass
    return engines


# TODO: Deprecate/remove this.
def get_default_engine(config, section='rattail.db'):
    """
    Fetch the default database engine defined in the given config object for a
    given section.

    :param config: A ``ConfigParser`` instance containing app configuration.

    :type section: string
    :param section: Optional section name within which the configuration
       options are defined.  If not specified, ``'rattail.db'`` is assumed.

    :returns: A SQLAlchemy engine instance, or ``None``.

    .. note::
       This function calls :func:`get_engines()` for the heavy lifting; see
       that function for more details on how the engine configuration is read.
    """
    return get_engines(config, section=section).get('default')


def configure_session(config, session):
    """
    Configure a session factory or instance.  Currently all this does is
    install the hook to record changes, if so config so dictates.
    """
    from rattail.db.changes import record_changes

    if config.getbool('rattail.db', 'changes.record', usedb=False):
        ignore_role_changes = config.getbool(
            'rattail.db', 'changes.ignore_roles', default=True, usedb=False)
        record_changes(session, ignore_role_changes)


def configure_versioning(config, manager=None, plugins=None, **kwargs):
    """
    Configure Continuum versioning.
    """
    if not config.getbool('rattail.db', 'versioning.enabled', default=False, usedb=False):
        return

    try:
        import sqlalchemy_continuum as continuum
        from rattail.db.continuum import versioning_manager, RattailPlugin
    except ImportError as error:
        raise SQLAlchemyNotInstalled(error)
    else:
        if manager is None:
            manager = versioning_manager
        if plugins is None:
            plugins = [RattailPlugin()]
        log.info("enabling Continuum versioning")
        continuum.make_versioned(manager=manager, plugins=plugins, **kwargs)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
File Utilities

This module contains various utility functions for use with the filesystem.
"""

from __future__ import unicode_literals

import io
import os
import shutil
import lockfile
import tempfile
import errno
import warnings
from datetime import datetime

import pkg_resources

from rattail.exceptions import PathNotFound


def count_lines(path, encoding=None):
    """
    Counts the number of lines in a text file.  Some attempt is made to ensure
    cross-platform compatibility.

    :param path: Path to the file.
    :type path: string

    :param encoding: Optional encoding to be used when opening the file for
      reading.  This is passed directly to :func:`python:io.open()`.

    :returns: Number of lines in the file.
    :rtype: integer
    """
    with io.open(path, 'rt', encoding=encoding) as f:
        lines = f.read().count('\n')
    return lines


def creation_time(path):
    """
    Returns the "naive" (i.e. not timezone-aware) creation timestamp for a
    file.

    :param path: Path to the file.
    :type path: string

    :returns: The creation timestamp for the file.
    :rtype: ``datetime.datetime`` instance
    """

    time = os.path.getctime(path)
    return datetime.fromtimestamp(time)


def locking_copy(src, destdir, timeout=None):
    """
    Implements a "locking" version of the standard library's
    :func:`python:shutil.copy()` function.

    This exists to provide a more atomic method for copying a file into a
    folder which is being watched by a file monitor.  The assumption is that
    the monitor is configured to watch for file "locks" and therefore only
    process files once they have had their locks removed.  See also
    :ref:`filemon-profile-watchlocks`.

    :param src: Path to the source file.
    :type src: string

    :param destdir: Path to the destination directory.
    :type destdir: string
    """
    # TODO: Remove the timeout arg altogether.
    if timeout is not None:
        warnings.warn("The 'timeout' arg is deprecated and will be ignored.",
                      DeprecationWarning)

    fn = os.path.basename(src)
    dst = os.path.join(destdir, fn)
    lockdir = '{0}.lock'.format(dst)

    # Note that we normalize two separate errors to PathNotFound below.  This
    # is because the file monitor retry mechanism will fail if two distinct
    # error types are encountered.  Since this function needs to be somewhat
    # robust in the face of network problems, raising a common error type will
    # give us a better chance at being retried (if so configured).

    # Attempt to create the lock dir.  We ignore "file exists" error here since
    # it's possible this function may be called mutliple times as part of a
    # filemon retry strategy.
    try:
        os.mkdir(lockdir)
    except OSError as error:
        if error.errno == errno.ENOENT:
            raise PathNotFound(error)
        if error.errno != errno.EEXIST:
            raise

    # Attempt to copy the file.
    try:
        shutil.copy(src, dst)
    except IOError as error:
        if error.errno == errno.ENOENT:
            raise PathNotFound(error)
        raise

    # Attempt to remove the lock dir.
    os.rmdir(lockdir)


def locking_copy_old(src, dst, timeout=None):
    """
    Implements a "locking" version of the standard library's
    :func:`python:shutil.copy()` function.

    This exists to provide a more atomic method for copying a file into a
    folder which is being watched by a file monitor.  The assumption is that
    the monitor is configured to watch for file "locks" and therefore only
    process files once they have had their locks removed.  See also
    :ref:`filemon-profile-watchlocks`.

    :param src: Path to the source file.
    :type src: string

    :param dst: Path to the destination file (or directory).
    :type dst: string

    :type timeout: float
    :param timeout: Number of seconds to wait for the file lock to clear, if it
       already exists.  This value may be specified as an integer or float, or
       string (which will be coerced to a float).

       .. note::
          There is no default value for the timeout, which means that by
          default, the function will wait indefinitely for the lock to clear.
    """
    # Coerce timeout to float in case it isn't already, e.g. in the case of
    # being called as a filemon action.
    if timeout is not None:
        timeout = float(timeout)

    if os.path.isdir(dst):
        fn = os.path.basename(src)
        dst = os.path.join(dst, fn)

    with lockfile.LockFile(dst, timeout=timeout):
        shutil.copy(src, dst)


def overwriting_move(src, dst):
    """
    Convenience function which is equivalent to ``shutil.move()``, except it
    will cause the destination file to be overwritten if it exists.
    """

    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    if os.path.exists(dst):
        os.remove(dst)
    shutil.move(src, dst)


def move_lpt(src, dst):
    """
    "Move" a source file to a LPT port, on Windows only.

    This is an experimental function which will hopefully provide a more
    reliable way to send a print job directly to a parallel port (e.g. 'LPT1')
    on Windows systems.  The function body was copied from
    :func:`python:shutil.move()` and "gutted" as much as was practical.

    Note that this function's signature differs from ``shutil.move()`` in that
    the ``dst`` parameter is always assumed to be a LPT port, e.g. ``'LPT1'``,
    and never a "file" path.
    """
    try:
        os.rename(src, dst)
    except OSError:
        shutil.copyfile(src, dst)
        os.unlink(src)


def resource_path(path):
    """
    Obtain a resource file path, extracting the resource and/or coercing the
    path as necessary.

    :param path: May be either a package resource specifier, or a regular file
       path.
    :type path: string

    :returns: Absolute file path to the resource.
    :rtype: string

    If ``path`` is a package resource specifier, and the package containing it
    is a zipped egg, then the resource will be extracted and the resultant
    filename will be returned.
    """

    if not os.path.isabs(path) and ':' in path:
        return pkg_resources.resource_filename(*path.split(':'))
    return path


def temp_path(suffix='.tmp', prefix='rattail.', **kwargs):
    """
    Returns a path for a temporary file.

    This is a convenience function which wraps ``tempfile.mkstemp()``.  The
    meaning of the arguments is the same.
    """

    fd, path = tempfile.mkstemp(suffix, prefix, **kwargs)
    os.close(fd)
    return path



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Import Handlers
"""

from __future__ import unicode_literals, absolute_import

import sys
import datetime
import logging

import humanize

from rattail.util import OrderedDict
from rattail.mail import send_email


log = logging.getLogger(__name__)


class ImportHandler(object):
    """
    Base class for all import handlers.
    """
    local_title = "Rattail"
    host_title = "Host/Other"
    session = None
    progress = None
    dry_run = False

    def __init__(self, config=None, **kwargs):
        self.config = config
        self.importers = self.get_importers()
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

    def get_importers(self):
        """
        Returns a dict of all available importers, where the values are
        importer factories.  All subclasses will want to override this.  Note
        that if you return an ``OrderedDict`` instance, you can affect the
        ordering of keys in the command line help system, etc.
        """
        return {}

    def get_importer_keys(self):
        """
        Returns a list of keys corresponding to the available importers.
        """
        return list(self.importers.iterkeys())

    def get_default_keys(self):
        """
        Returns a list of keys corresponding to the default importers.
        Override this if you wish certain importers to be excluded by default,
        e.g. when first testing them out etc.
        """
        return self.get_importer_keys()

    def get_importer(self, key):
        """
        Returns an importer instance corresponding to the given key.
        """
        kwargs = self.get_importer_kwargs(key)
        kwargs['config'] = self.config
        kwargs['session'] = self.session
        importer = self.importers[key](**kwargs)
        importer.handler = self
        return importer

    def get_importer_kwargs(self, key):
        """
        Return a dict of kwargs to be used when construcing an importer with
        the given key.
        """
        kwargs = {}
        if hasattr(self, 'host_session'):
            kwargs['host_session'] = self.host_session
        return kwargs

    def import_data(self, keys, args):
        """
        Import all data for the given importer keys.
        """
        self.now = datetime.datetime.utcnow()
        self.dry_run = args.dry_run
        self.begin_transaction()
        self.setup()
        changes = OrderedDict()

        for key in keys:
            importer = self.get_importer(key)
            if not importer:
                log.warning("skipping unknown importer: {}".format(key))
                continue

            created, updated, deleted = importer.import_data(args, progress=self.progress)

            changed = bool(created or updated or deleted)
            logger = log.warning if changed and args.warnings else log.info
            logger("{} -> {}: added {}, updated {}, deleted {} {} records".format(
                self.host_title, self.local_title, len(created), len(updated), len(deleted), key))
            if changed:
                changes[key] = created, updated, deleted

        if changes:
            self.process_changes(changes, args)

        if self.dry_run:
            self.rollback_transaction()
        else:
            self.commit_transaction()

        self.teardown()
        return changes

    def begin_transaction(self):
        self.begin_host_transaction()
        self.begin_local_transaction()

    def begin_host_transaction(self):
        if hasattr(self, 'make_host_session'):
            self.host_session = self.make_host_session()

    def begin_local_transaction(self):
        pass

    def setup(self):
        """
        Perform any setup necessary, prior to running the import task(s).
        """

    def rollback_transaction(self):
        self.rollback_host_transaction()
        self.rollback_local_transaction()

    def rollback_host_transaction(self):
        if hasattr(self, 'host_session'):
            self.host_session.rollback()
            self.host_session.close()
            self.host_session = None

    def rollback_local_transaction(self):
        pass

    def commit_transaction(self):
        self.commit_host_transaction()
        self.commit_local_transaction()

    def commit_host_transaction(self):
        if hasattr(self, 'host_session'):
            self.host_session.commit()
            self.host_session.close()
            self.host_session = None

    def commit_local_transaction(self):
        pass

    def teardown(self):
        """
        Perform any cleanup necessary, after running the import task(s).
        """

    def process_changes(self, changes, args):
        """
        This method is called any time changes occur, regardless of whether the
        import is running in "warnings" mode.  Default implementation however
        is to do nothing unless warnings mode is in effect, in which case an
        email notification will be sent.
        """
        # TODO: This whole thing needs a re-write...but for now, waiting until
        # the old importer has really gone away, so we can share its email
        # template instead of bothering with something more complicated.

        if not args.warnings:
            return

        data = {
            'local_title': self.local_title,
            'host_title': self.host_title,
            'argv': sys.argv,
            'runtime': humanize.naturaldelta(datetime.datetime.utcnow() - self.now),
            'changes': changes,
            'dry_run': args.dry_run,
            'render_record': RecordRenderer(self.config),
            'max_display': 15,
        }

        command = getattr(self, 'command', None)
        if command:
            data['command'] = '{} {}'.format(command.parent.name, command.name)
        else:
            data['command'] = None

        if command:
            key = '{}_{}_updates'.format(command.parent.name, command.name)
            key = key.replace('-', '_')
        else:
            key = 'rattail_import_updates'

        send_email(self.config, key, fallback_key='rattail_import_updates', data=data)


class SQLAlchemyImportHandler(ImportHandler):
    """
    Handler for imports for which the host data source is represented by a
    SQLAlchemy engine and ORM.
    """
    host_session = None

    def make_host_session(self):
        raise NotImplementedError


class BulkPostgreSQLImportHandler(ImportHandler):
    """
    Handler for bulk imports which target PostgreSQL on the local side.
    """

    def import_data(self, keys, args):
        """
        Import all data for the given importer keys.
        """
        self.now = datetime.datetime.utcnow()
        self.dry_run = args.dry_run
        self.begin_transaction()
        self.setup()

        for key in keys:
            importer = self.get_importer(key)
            if not importer:
                log.warning("skipping unknown importer: {}".format(key))
                continue

            created = importer.import_data(args, progress=self.progress)
            log.info("{} -> {}: added {}, updated 0, deleted 0 {} records".format(
                self.host_title, self.local_title, created, key))

        if self.dry_run:
            self.rollback_transaction()
        else:
            self.commit_transaction()

        self.teardown()


class RecordRenderer(object):
    """
    Record renderer for email notifications sent from data import jobs.
    """

    def __init__(self, config):
        self.config = config

    def __call__(self, record):
        return self.render(record)

    def render(self, record):
        """
        Render the given record.
        """
        key = record.__class__.__name__.lower()
        renderer = getattr(self, 'render_{}'.format(key), None)
        if renderer:
            return renderer(record)

        label = self.get_label(record)
        url = self.get_url(record)
        if url:
            return '<a href="{}">{}</a>'.format(url, label)
        return label

    def get_label(self, record):
        key = record.__class__.__name__.lower()
        label = getattr(self, 'label_{}'.format(key), self.label)
        return label(record)

    def label(self, record):
        return unicode(record)

    def get_url(self, record):
        """
        Fetch / generate a URL for the given data record.  You should *not*
        override this method, but do :meth:`url()` instead.
        """
        key = record.__class__.__name__.lower()
        url = getattr(self, 'url_{}'.format(key), self.url)
        return url(record)

    def url(self, record):
        """
        Fetch / generate a URL for the given data record.
        """
        if hasattr(record, 'uuid'):
            url = self.config.get('tailbone', 'url')
            if url:
                url = url.rstrip('/')
                name = '{}s'.format(record.__class__.__name__.lower())
                if name == 'persons': # FIXME, obviously this is a hack
                    name = 'people'
                url = '{}/{}/{{uuid}}'.format(url, name)
                return url.format(uuid=record.uuid)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for *OLD* Batch System

Please don't use any of this for future development!  New batches should
instead be based on :class:`rattail.db.batch.BatchMixin` and friends.
"""

from __future__ import unicode_literals

import re
import datetime
import logging

import sqlalchemy as sa
from sqlalchemy.orm import relationship, object_session
from sqlalchemy.ext.orderinglist import ordering_list

from .core import Base, uuid_column, GPCType
from rattail.sil import get_column
 

log = logging.getLogger(__name__)


class Batch(Base):
    """
    Represents a SIL-compliant batch of data.
    """

    __tablename__ = 'batch'

    uuid = uuid_column()
    provider = sa.Column(sa.String(length=50))
    id = sa.Column(sa.String(length=8))
    source = sa.Column(sa.String(length=6))
    destination = sa.Column(sa.String(length=6))
    action_type = sa.Column(sa.String(length=6))
    description = sa.Column(sa.String(length=50))
    rowcount = sa.Column(sa.Integer(), default=0)
    executed = sa.Column(sa.DateTime())
    # TODO: Convert this to a DateTime, to handle time zone issues.
    purge = sa.Column(sa.Date())

    _rowclasses = {}

    sil_type_pattern = re.compile(r'^(CHAR|NUMBER)\((\d+(?:\,\d+)?)\)$')

    def __unicode__(self):
        return unicode(self.description or '')

    @property
    def rowclass(self):
        """
        Returns the mapped class for the underlying row (data) table.
        """

        if not self.uuid:
            object_session(self).flush()
            assert self.uuid

        if self.uuid not in self._rowclasses:

            kwargs = {
                '__tablename__': 'batch.%s' % self.uuid,
                'uuid': uuid_column(),
                'ordinal': sa.Column(sa.Integer(), nullable=False),
                }

            for column in self.columns:
                data_type = self.get_sqlalchemy_type(column.data_type)
                kwargs[column.name] = sa.Column(data_type)
            rowclass = type('BatchRow_{0}'.format(self.uuid).encode('ascii'), (Base, BatchRow), kwargs)

            batch_uuid = self.uuid
            def batch(self):
                return object_session(self).query(Batch).get(batch_uuid)
            rowclass.batch = property(batch)

            self._rowclasses[self.uuid] = rowclass

        return self._rowclasses[self.uuid]

    @classmethod
    def get_sqlalchemy_type(cls, sil_type):
        """
        Returns a SQLAlchemy data type according to a SIL data type.
        """

        if sil_type == 'GPC(14)':
            return GPCType()

        if sil_type == 'FLAG(1)':
            return sa.Boolean()

        m = cls.sil_type_pattern.match(sil_type)
        if m:
            data_type, precision = m.groups()
            if precision.isdigit():
                precision = int(precision)
                scale = 0
            else:
                precision, scale = precision.split(',')
                precision = int(precision)
                scale = int(scale)
            if data_type == 'CHAR':
                assert not scale, "FIXME"
                return sa.String(length=precision)
            if data_type == 'NUMBER':
                return sa.Numeric(precision=precision, scale=scale)

        assert False, "FIXME"

    def add_column(self, sil_name=None, **kwargs):
        column = BatchColumn(sil_name, **kwargs)
        self.columns.append(column)

    def add_row(self, row, **kwargs):
        """
        Adds a row to the batch data table.
        """
        session = object_session(self)
        # FIXME: This probably needs to use a func.max() query.
        row.ordinal = self.rowcount + 1
        session.add(row)
        self.rowcount += 1
        session.flush()

    def create_table(self):
        """
        Creates the batch's data table within the database.
        """
        session = object_session(self)
        self.rowclass.__table__.create(session.bind)

    def drop_table(self):
        """
        Drops the batch's data table from the database.
        """
        log.debug("dropping normal batch table: {0}".format(self.rowclass.__table__.name))
        session = object_session(self)
        self.rowclass.__table__.drop(bind=session.bind, checkfirst=True)

    def execute(self, config, progress=None):
        from rattail.batches.exceptions import BatchProviderNotFound

        try:
            provider = self.get_provider(config)
            if not provider.execute(self, progress):
                return False

        except BatchProviderNotFound:
            executor = self.get_executor(config)
            if not executor.execute(self, progress):
                return False

        self.executed = datetime.datetime.utcnow()
        object_session(self).flush()
        return True

    def get_provider(self, config):
        from rattail.batches import get_provider
        assert self.provider
        return get_provider(config, self.provider)

    def get_executor(self, config):
        from rattail.db.batches import get_batch_executor
        assert self.provider
        return get_batch_executor(self.provider, config)

    @property
    def rows(self):
        session = object_session(self)
        q = session.query(self.rowclass)
        q = q.order_by(self.rowclass.ordinal)
        return q


class BatchColumn(Base):
    """
    Represents a SIL column associated with a batch.
    """
    __tablename__ = 'batch_column'
    __table_args__ = (
        sa.ForeignKeyConstraint(['batch_uuid'], ['batch.uuid'], name='batch_column_fk_batch'),
        )

    uuid = uuid_column()
    batch_uuid = sa.Column(sa.String(length=32))
    ordinal = sa.Column(sa.Integer(), nullable=False)
    name = sa.Column(sa.String(length=20))
    display_name = sa.Column(sa.String(length=50))
    sil_name = sa.Column(sa.String(length=10))
    data_type = sa.Column(sa.String(length=15))
    description = sa.Column(sa.String(length=50))
    visible = sa.Column(sa.Boolean(), default=True)

    def __init__(self, sil_name=None, **kwargs):
        if sil_name:
            kwargs['sil_name'] = sil_name
            sil_column = get_column(sil_name)
            kwargs.setdefault('name', sil_name)
            kwargs.setdefault('data_type', sil_column.data_type)
            kwargs.setdefault('description', sil_column.description)
            kwargs.setdefault('display_name', sil_column.display_name)
        super(BatchColumn, self).__init__(**kwargs)

    def __unicode__(self):
        return unicode(self.display_name or '')


Batch.columns = relationship(
    BatchColumn, backref='batch',
    collection_class=ordering_list('ordinal'),
    order_by=BatchColumn.ordinal,
    cascade='save-update, merge, delete, delete-orphan')


class BatchRow(object):
    """
    Superclass of batch row objects.
    """

    def __unicode__(self):
        return u"Row {0}".format(self.ordinal)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch Utilities
"""

from __future__ import unicode_literals

import re
import datetime
import logging

from sqlalchemy import MetaData
from sqlalchemy import and_

from .. import Session
from .. import model


batch_pattern = re.compile(r'^batch\.[0-9a-f]{32}$')

log = logging.getLogger(__name__)


def purge_batches(effective_date=None, purge_everything=False):
    """
    Purge old batches from the database.

    :param effective_date: Date against which comparisons should be made when
      determining if a batch is "old" (based on its ``purge_date`` attribute).
      The current date is assumed if none is specified.

    :param purge_everything: Boolean indicating whether the purge should be
      complete and indiscriminate, as opposed to honoring batch purge dates.
      (Use with caution!)

    :returns: Number of batches purged.
    :rtype: int
    """
    if effective_date is None:
        effective_date = datetime.date.today()

    session = Session()
    batches = session.query(model.Batch)
    if not purge_everything:
        batches = batches.filter(and_(
                model.Batch.purge != None,
                model.Batch.purge < effective_date))

    purged = 0
    for batch in batches:
        batch.drop_table()
        session.delete(batch)
        purged += 1
    session.commit()
    session.close()
    return purged


def purge_orphaned_batches():
    """
    Drop any orphaned batch tables which happen to still exist.

    This should theoretically not be necessary, if/when the batch processing is
    cleaning up after itself properly.  For now though, it seems that orphaned
    data tables are sometimes being left behind.  This removes them.
    """
    session = Session()
    current_batches = []
    for batch in session.query(model.Batch):
        current_batches.append('batch.{0}'.format(batch.uuid))
    session.close()

    def orphaned_batches(name, metadata):
        return batch_pattern.match(name) and name not in current_batches

    metadata = MetaData(session.bind)
    metadata.reflect(only=orphaned_batches)
    count = len(metadata.tables)
    for table in reversed(metadata.sorted_tables):
        log.debug("dropping orphaned batch table: {0}".format(table.name))
        table.drop()
    return count



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail Data Import
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import Session, newimporting as importing
from rattail.util import OrderedDict


class RattailImportHandler(importing.SQLAlchemyImportHandler):
    """
    Handler for Rattail -> Rattail data import.
    """
    host_title = "Rattail"
    dbkey = 'host'

    def make_host_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])

    def get_importers(self):
        importers = OrderedDict()
        importers['Person'] = PersonImporter
        importers['PersonEmailAddress'] = PersonEmailAddressImporter
        importers['PersonPhoneNumber'] = PersonPhoneNumberImporter
        importers['PersonMailingAddress'] = PersonMailingAddressImporter
        importers['User'] = UserImporter
        importers['Message'] = MessageImporter
        importers['MessageRecipient'] = MessageRecipientImporter
        importers['Store'] = StoreImporter
        importers['StorePhoneNumber'] = StorePhoneNumberImporter
        importers['Employee'] = EmployeeImporter
        importers['EmployeeStore'] = EmployeeStoreImporter
        importers['EmployeeDepartment'] = EmployeeDepartmentImporter
        importers['EmployeeEmailAddress'] = EmployeeEmailAddressImporter
        importers['EmployeePhoneNumber'] = EmployeePhoneNumberImporter
        importers['ScheduledShift'] = ScheduledShiftImporter
        importers['WorkedShift'] = WorkedShiftImporter
        importers['Customer'] = CustomerImporter
        importers['CustomerGroup'] = CustomerGroupImporter
        importers['CustomerGroupAssignment'] = CustomerGroupAssignmentImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['CustomerEmailAddress'] = CustomerEmailAddressImporter
        importers['CustomerPhoneNumber'] = CustomerPhoneNumberImporter
        importers['Vendor'] = VendorImporter
        importers['VendorEmailAddress'] = VendorEmailAddressImporter
        importers['VendorPhoneNumber'] = VendorPhoneNumberImporter
        importers['VendorContact'] = VendorContactImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Category'] = CategoryImporter
        importers['Family'] = FamilyImporter
        importers['ReportCode'] = ReportCodeImporter
        importers['DepositLink'] = DepositLinkImporter
        importers['Tax'] = TaxImporter
        importers['Brand'] = BrandImporter
        importers['Product'] = ProductImporter
        importers['ProductCode'] = ProductCodeImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductPrice'] = ProductPriceImporter
        return importers

    def get_importer_kwargs(self, key):
        kwargs = super(RattailImportHandler, self).get_importer_kwargs(key)
        kwargs['dbkey'] = self.dbkey
        return kwargs


class RattailImporter(importing.SQLAlchemyImporter):
    """
    Base class for Rattail -> Rattail data importers.
    """

    @property
    def host_model_class(self):
        return self.model_class

    @property
    def supported_fields(self):
        """
        We only need to support the simple fields in a Rattail->Rattail import,
        since all relevant tables should be covered and therefore no need to do
        crazy foreign key acrobatics etc.
        """
        return self.simple_fields

    @property
    def normalize_progress_message(self):
        return "Reading {} data from '{}' db".format(self.model_name, self.dbkey)

    def query(self):
        query = super(RattailImporter, self).query()
        options = self.cache_query_options()
        if options:
            for option in options:
                query = query.options(option)
        return query

    def normalize_source_record(self, host_instance):
        return self.normalize_instance(host_instance)


class PersonImporter(RattailImporter, importing.model.PersonImporter):
    pass

class PersonEmailAddressImporter(RattailImporter, importing.model.PersonEmailAddressImporter):
    pass

class PersonPhoneNumberImporter(RattailImporter, importing.model.PersonPhoneNumberImporter):
    pass

class PersonMailingAddressImporter(RattailImporter, importing.model.PersonMailingAddressImporter):
    pass

class UserImporter(RattailImporter, importing.model.UserImporter):
    pass

class MessageImporter(RattailImporter, importing.model.MessageImporter):
    pass

class MessageRecipientImporter(RattailImporter, importing.model.MessageRecipientImporter):
    pass

class StoreImporter(RattailImporter, importing.model.StoreImporter):
    pass

class StorePhoneNumberImporter(RattailImporter, importing.model.StorePhoneNumberImporter):
    pass

class EmployeeImporter(RattailImporter, importing.model.EmployeeImporter):
    pass

class EmployeeStoreImporter(RattailImporter, importing.model.EmployeeStoreImporter):
    pass

class EmployeeDepartmentImporter(RattailImporter, importing.model.EmployeeDepartmentImporter):
    pass

class EmployeeEmailAddressImporter(RattailImporter, importing.model.EmployeeEmailAddressImporter):
    pass

class EmployeePhoneNumberImporter(RattailImporter, importing.model.EmployeePhoneNumberImporter):
    pass

class ScheduledShiftImporter(RattailImporter, importing.model.ScheduledShiftImporter):
    pass

class WorkedShiftImporter(RattailImporter, importing.model.WorkedShiftImporter):
    pass

class CustomerImporter(RattailImporter, importing.model.CustomerImporter):
    pass

class CustomerGroupImporter(RattailImporter, importing.model.CustomerGroupImporter):
    pass

class CustomerGroupAssignmentImporter(RattailImporter, importing.model.CustomerGroupAssignmentImporter):
    pass

class CustomerPersonImporter(RattailImporter, importing.model.CustomerPersonImporter):
    pass

class CustomerEmailAddressImporter(RattailImporter, importing.model.CustomerEmailAddressImporter):
    pass

class CustomerPhoneNumberImporter(RattailImporter, importing.model.CustomerPhoneNumberImporter):
    pass

class VendorImporter(RattailImporter, importing.model.VendorImporter):
    pass

class VendorEmailAddressImporter(RattailImporter, importing.model.VendorEmailAddressImporter):
    pass

class VendorPhoneNumberImporter(RattailImporter, importing.model.VendorPhoneNumberImporter):
    pass

class VendorContactImporter(RattailImporter, importing.model.VendorContactImporter):
    pass

class DepartmentImporter(RattailImporter, importing.model.DepartmentImporter):
    pass

class SubdepartmentImporter(RattailImporter, importing.model.SubdepartmentImporter):
    pass

class CategoryImporter(RattailImporter, importing.model.CategoryImporter):
    pass

class FamilyImporter(RattailImporter, importing.model.FamilyImporter):
    pass

class ReportCodeImporter(RattailImporter, importing.model.ReportCodeImporter):
    pass

class DepositLinkImporter(RattailImporter, importing.model.DepositLinkImporter):
    pass

class TaxImporter(RattailImporter, importing.model.TaxImporter):
    pass

class BrandImporter(RattailImporter, importing.model.BrandImporter):
    pass

class ProductImporter(RattailImporter, importing.model.ProductImporter):
    pass

class ProductCodeImporter(RattailImporter, importing.model.ProductCodeImporter):
    pass

class ProductCostImporter(RattailImporter, importing.model.ProductCostImporter):
    pass

class ProductPriceImporter(RattailImporter, importing.model.ProductPriceImporter):
    pass



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail bulk data import
"""

from __future__ import unicode_literals, absolute_import

import os

from rattail.db import Session, newimporting as importing
from rattail.util import OrderedDict


class RattailImportHandler(importing.BulkPostgreSQLImportHandler):
    """
    Handler for Rattail -> Rattail bulk data import.
    """
    host_title = "Rattail"
    dbkey = 'host'

    def make_host_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])

    def get_importers(self):
        importers = OrderedDict()
        importers['ScheduledShift'] = ScheduledShiftImporter
        importers['WorkedShift'] = WorkedShiftImporter
        return importers

    def get_importer_kwargs(self, key):
        kwargs = super(RattailImportHandler, self).get_importer_kwargs(key)
        kwargs['dbkey'] = self.dbkey
        return kwargs


class RattailImporter(importing.SQLAlchemyImporter, importing.BulkPostgreSQLImporter):
    """
    Base class for Rattail -> Rattail bulk data importers.
    """

    @property
    def host_model_class(self):
        return self.model_class

    @property
    def supported_fields(self):
        """
        We only need to support the simple fields in a Rattail->Rattail import,
        since all relevant tables should be covered and therefore no need to do
        crazy foreign key acrobatics etc.
        """
        return self.simple_fields

    @property
    def data_path(self):
        return os.path.join(self.config.require('rattail', 'workdir'),
                            'import_{}.csv'.format(self.model_table.name))

    @property
    def normalize_progress_message(self):
        return "Reading {} data from '{}' db".format(self.model_name, self.dbkey)

    def query(self):
        query = super(RattailImporter, self).query()
        options = self.cache_query_options()
        if options:
            for option in options:
                query = query.options(option)
        return query

    def normalize_source_record(self, host_object):
        return self.normalize_instance(host_object)


class ScheduledShiftImporter(RattailImporter, importing.model.ScheduledShiftImporter):
    pass

class WorkedShiftImporter(RattailImporter, importing.model.WorkedShiftImporter):
    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from ... import DataTestCase

from rattail.db import model


class TestUserEmailAddress(DataTestCase):

    def extra_setup(self):
        self.user = model.User(username='fred')
        self.session.add(self.user)
        self.session.flush()

    def test_email_defaults_to_none(self):
        self.assertTrue(self.user.get_email_address() is None)

    def test_email_comes_from_person_then_customer(self):
        # only customer has email at this point
        person = model.Person(first_name='Fred')
        customer = model.Customer(name='Fred')
        customer.add_email_address('customer@mailinator.com')
        customer.people.append(person)
        self.user.person = person
        self.session.flush()
        self.assertEqual(self.user.get_email_address(), 'customer@mailinator.com')

        # now person email will take preference
        person.add_email_address('person@mailinator.com')
        self.session.refresh(person)
        self.assertEqual(self.user.get_email_address(), 'person@mailinator.com')

    def test_email_address_property_works_too(self):
        # even though this may go away some day, cover it for now
        person = model.Person(first_name='Fred')
        person.add_email_address('person@mailinator.com')
        self.user.person = person
        self.session.flush()
        self.assertEqual(self.user.email_address, 'person@mailinator.com')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Utilities
"""

from __future__ import unicode_literals

import sys

from pkg_resources import iter_entry_points

try:
    from collections import OrderedDict
except ImportError: # pragma no cover
    from ordereddict import OrderedDict


def import_module_path(module_path):
    """
    Import an arbitrary Python module.

    :param module_path: String referencing a module by its "dotted path".

    :returns: The referenced module.
    """
    if module_path in sys.modules:
        return sys.modules[module_path]
    module = __import__(module_path)

    def last_module(module, module_path):
        parts = module_path.split('.')
        parts.pop(0)
        child = getattr(module, parts[0])
        if len(parts) == 1:
            return child
        return last_module(child, '.'.join(parts))

    return last_module(module, module_path)


def load_object(specifier):
    """
    Load an arbitrary object from a module, according to a specifier.

    The specifier string should contain a dotted path to an importable module,
    followed by a colon (``':'``), followed by the name of the object to be
    loaded.  For example:

    .. code-block:: none

       rattail.files:overwriting_move

    You'll notice from this example that "object" in this context refers to any
    valid Python object, i.e. not necessarily a class instance.  The name may
    refer to a class, function, variable etc.  Once the module is imported, the
    ``getattr()`` function is used to obtain a reference to the named object;
    therefore anything supported by that method should work.

    :param specifier: Specifier string.

    :returns: The specified object.
    """
    module_path, name = specifier.split(':')
    module = import_module_path(module_path)
    return getattr(module, name)


def load_entry_points(group):
    """
    Load a set of ``setuptools``-style entry points.

    This is a convenience wrapper around ``pkg_resources.iter_entry_points()``.

    :param group: The group of entry points to be loaded.

    :returns: A dictionary whose keys are the entry point names, and values are
       the loaded entry points.
    """
    entry_points = {}
    for entry_point in iter_entry_points(group):
        entry_points[entry_point.name] = entry_point.load()
    return entry_points



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model Importers
"""

from __future__ import unicode_literals, absolute_import

import logging

from sqlalchemy import orm

from rattail.db import model
from rattail.db.api import set_regular_price, set_current_sale_price
from rattail.db.auth import administrator_role
from rattail.db.importing import Importer, normal


log = logging.getLogger(__name__)


class PersonImporter(Importer):
    """
    Person data importer.
    """
    model_class = model.Person
    simple_fields = [
        'uuid',
        'first_name',
        'middle_name',
        'last_name',
        'display_name',
    ]
    supported_fields = simple_fields


class PersonEmailAddressImporter(Importer):
    """
    Person email address data importer.
    """
    model_class = model.PersonEmailAddress
    simple_fields = [
        'uuid',
        'parent_uuid',
        'type',
        'address',
        'preference',
        'invalid',
    ]
    supported_fields = simple_fields + [
        'preferred',
        ]

    def normalize_instance(self, email):
        data = super(PersonEmailAddressImporter, self).normalize_instance(email)
        if 'preferred' in self.fields:
            data['preferred'] = email.preference == 1
        return data

    def update_instance(self, email, data, inst_data=None):
        super(PersonEmailAddressImporter, self).update_instance(email, data, inst_data)
        if 'preferred' in self.fields:
            if data['preferred']:
                if email.preference != 1:
                    person = email.person
                    if not person:
                        person = self.session.query(model.Person).get(email.parent_uuid)
                        assert person, "hm: " + email.parent_uuid
                    if email in person.emails:
                        person.emails.remove(email)
                    person.emails.insert(0, email)
                    person.emails.reorder()
            else:
                if email.preference == 1:
                    person = email.person
                    if not person:
                        person = self.session.query(model.Person).get(email.parent_uuid)
                        assert person, "hm: " + email.parent_uuid
                    if len(person.emails) > 1:
                        person.emails.remove(email)
                        person.emails.append(email)
                        person.emails.reorder()

        # If this is a new record, we may still need to establish its preference.
        if email.preference is None:
            person = email.person
            if not person:
                person = self.session.query(model.Person).get(email.parent_uuid)
                assert person, "hm: " + email.parent_uuid
            if email not in person.emails:
                person.emails.append(email)
            person.emails.reorder()


class PersonPhoneNumberImporter(Importer):
    """
    Person phone number data importer.
    """
    model_class = model.PersonPhoneNumber
    simple_fields = [
        'uuid',
        'parent_uuid',
        'type',
        'number',
        'preference',
        ]
    supported_fields = simple_fields + [
        'preferred',
        ]

    def normalize_instance(self, phone):
        data = super(PersonPhoneNumberImporter, self).normalize_instance(phone)
        if 'preferred' in self.fields:
            data['preferred'] = phone.preference == 1
        return data

    def update_instance(self, phone, data, inst_data=None):
        super(PersonPhoneNumberImporter, self).update_instance(phone, data, inst_data)
        if 'preferred' in self.fields:
            if data['preferred']:
                if phone.preference != 1:
                    person = phone.person
                    if not person:
                        person = self.session.query(model.Person).get(phone.parent_uuid)
                        assert person, "hm: " + phone.parent_uuid
                    if phone in person.phones:
                        person.phones.remove(phone)
                    person.phones.insert(0, phone)
                    person.phones.reorder()
            else:
                if phone.preference == 1:
                    person = phone.person
                    if not person:
                        person = self.session.query(model.Person).get(phone.parent_uuid)
                        assert person, "hm: " + phone.parent_uuid
                    if len(person.phones) > 1:
                        person.phones.remove(phone)
                        person.phones.append(phone)
                        person.phones.reorder()

        # If this is a new record, we may still need to establish its preference.
        if phone.preference is None:
            person = phone.person
            if not person:
                person = self.session.query(model.Person).get(phone.parent_uuid)
                assert person, "hm: " + phone.parent_uuid
            if phone not in person.phones:
                person.phones.append(phone)
            person.phones.reorder()


class PersonMailingAddressImporter(Importer):
    """
    Person mailing address data importer.
    """
    model_class = model.PersonMailingAddress
    supported_fields = [
        'uuid',
        'parent_uuid',
        'type',
        'preference',
        'street',
        'street2',
        'city',
        'state',
        'zipcode',
        'invalid',
        ]


class UserImporter(Importer):
    """
    User data importer.
    """
    model_class = model.User
    normalizer_class = normal.UserNormalizer
    simple_fields = [
        'uuid',
        'username',
        'password',
        'salt',
        'person_uuid',
        'active',
    ]
    supported_fields = simple_fields + [
        'admin',
    ]

    def setup(self):
        self.normalizer.admin = self.admin = administrator_role(self.session)

    def update_instance(self, user, data, inst_data=None):
        super(UserImporter, self).update_instance(user, data, inst_data)
        if 'admin' in data:
            if data['admin']:
                if self.admin not in user.roles:
                    user.roles.append(self.admin)
            else:
                if self.admin in user.roles:
                    user.roles.remove(self.admin)


class MessageImporter(Importer):
    """
    User message data importer.
    """
    model_class = model.Message
    normalizer_class = normal.MessageNormalizer
    simple_fields = [
        'uuid',
        'sender_uuid',
        'subject',
        'body',
        'sent',
    ]


class MessageRecipientImporter(Importer):
    """
    User message recipient data importer.
    """
    model_class = model.MessageRecipient
    normalizer_class = normal.MessageRecipientNormalizer
    simple_fields = [
        'uuid',
        'message_uuid',
        'recipient_uuid',
        'status',
    ]


class StoreImporter(Importer):
    """
    Store data importer.
    """
    model_class = model.Store
    simple_fields = [
        'uuid',
        'id',
        'name',
        ]
    supported_fields = simple_fields + [
        'phone_number',
        'fax_number',
        ]

    def cache_query_options(self):
        if 'phone_number' in self.fields or 'fax_number' in self.fields:
            return [orm.joinedload(model.Store.phones)]

    def normalize_record(self, data):
        self.string_or_null(data, 'name', 'phone_number', 'fax_number')

    def normalize_instance(self, store):
        data = super(StoreImporter, self).normalize_instance(store)

        if 'phone_number' in self.fields:
            data['phone_number'] = None
            for phone in store.phones:
                if phone.type == 'Voice':
                    data['phone_number'] = phone.number
                    break

        if 'fax_number' in self.fields:
            data['fax_number'] = None
            for phone in store.phones:
                if phone.type == 'Fax':
                    data['fax_number'] = phone.number
                    break

        return data

    def update_instance(self, store, data, inst_data=None):
        super(StoreImporter, self).update_instance(store, data, inst_data)

        if 'phone_number' in data:
            number = data['phone_number'] or None
            if number:
                found = False
                for phone in store.phones:
                    if phone.type == 'Voice':
                        if phone.number != number:
                            phone.number = number
                        found = True
                        break
                if not found:
                    store.add_phone_number(number, type='Voice')
            else:
                for phone in list(store.phones):
                    if phone.type == 'Voice':
                        store.phones.remove(phone)

        if 'fax_number' in data:
            number = data['fax_number'] or None
            if number:
                found = False
                for phone in store.phones:
                    if phone.type == 'Fax':
                        if phone.number != number:
                            phone.number = number
                        found = True
                        break
                if not found:
                    store.add_phone_number(number, type='Fax')
            else:
                for phone in list(store.phones):
                    if phone.type == 'Fax':
                        store.phones.remove(phone)


class EmployeeImporter(Importer):
    """
    Employee data importer.
    """
    model_class = model.Employee
    normalizer_class = normal.EmployeeNormalizer
    simple_fields = [
        'uuid',
        'id',
        'person_uuid',
        'display_name',
        'status',
        ]
    supported_fields = simple_fields + [
        'customer_id',
        'first_name',
        'last_name',
        'person_display_name',
        'phone_number',
        'phone_number_2',
        'email_address',
        ]

    def setup(self):
        if 'customer_id' in self.fields:
            self.customers = self.cache_model(model.Customer, 'id')

    def cache_query_options(self):
        return [
            orm.joinedload(model.Employee.person).joinedload(model.Person._customers),
            orm.joinedload(model.Employee.phones),
            orm.joinedload(model.Employee.email),
        ]

    def normalize_record(self, data):
        self.string_or_null(data, 'customer_id', 'phone_number',
                            'phone_number_2', 'email_address')
        self.int_or_null(data, 'id', 'status')
        self.prioritize_2(data, 'phone_number')

    def update_instance(self, employee, data, inst_data=None):
        super(EmployeeImporter, self).update_instance(employee, data, inst_data)
        person = employee.person

        if 'first_name' in self.fields:
            employee.first_name = data['first_name']
        if 'last_name' in self.fields:
            employee.last_name = data['last_name']

        if 'person_display_name' in self.fields:
            if person.display_name != data['person_display_name']:
                person.display_name = data['person_display_name']

        if 'customer_id' in self.fields:
            id_ = data['customer_id']
            if id_:
                customer = self.customers.get(id_)
                if not customer:
                    customer = model.Customer()
                    customer.id = id_
                    customer.name = employee.display_name
                    self.session.add(customer)
                    self.customers[customer.id] = customer
                if person not in customer.people:
                    customer.people.append(person)
            else:
                for customer in list(person.customers):
                    if len(customer.people) > 1:
                        if person in customer.people:
                            customer.people.remove(person)

        if 'phone_number' in self.fields:
            number = data['phone_number']
            if number:
                found = False
                for phone in employee.phones:
                    if phone.type == 'Home':
                        if phone.number != number:
                            phone.number = number
                        found = True
                        break
                if not found:
                    employee.add_phone_number(number, type='Home')
            else:
                for phone in list(employee.phones):
                    if phone.type == 'Home':
                        employee.phones.remove(phone)

        if 'phone_number_2' in self.fields:
            number = data['phone_number_2']
            if number:
                found = False
                first = False
                for phone in employee.phones:
                    if phone.type == 'Home':
                        if first:
                            if phone.number != number:
                                phone.number = number
                            found = True
                            break
                        first = True
                if not found:
                    employee.add_phone_number(number, type='Home')
            else:
                first = False
                for phone in list(employee.phones):
                    if phone.type == 'Home':
                        if first:
                            employee.phones.remove(phone)
                            break
                        first = True

        if 'email_address' in self.fields:
            address = data['email_address']
            if address:
                if employee.email:
                    if employee.email.address != address:
                        employee.email.address = address
                else:
                    employee.add_email_address(address)
            else:
                if len(employee.emails):
                    del employee.emails[:]


class EmployeeStoreImporter(Importer):
    """
    Employee/store data importer.
    """
    model_class = model.EmployeeStore
    normalizer_class = normal.EmployeeStoreNormalizer
    simple_fields = [
        'uuid',
        'employee_uuid',
        'store_uuid',
    ]
    supported_fields = simple_fields


class EmployeeDepartmentImporter(Importer):
    """
    Employee/department data importer.
    """
    model_class = model.EmployeeDepartment
    normalizer_class = normal.EmployeeDepartmentNormalizer
    simple_fields = [
        'uuid',
        'employee_uuid',
        'department_uuid',
    ]
    supported_fields = simple_fields


class CustomerGroupImporter(Importer):
    """
    CustomerGroup data importer.
    """
    model_class = model.CustomerGroup
    supported_fields = [
        'uuid',
        'id',
        'name',
        ]

    def normalize_record(self, data):
        self.string_or_null(data, 'name')


class CustomerImporter(Importer):
    """
    Customer data importer.
    """
    model_class = model.Customer
    simple_fields = [
        'uuid',
        'id',
        'name',
        'email_preference',
        ]
    supported_fields = simple_fields + [
        'first_name',
        'last_name',
        'phone_number',
        'phone_number_2',
        'email_address',
        'group_id',
        'group_id_2',
        ]

    def setup(self):
        if 'group_id' in self.fields or 'group_id_2' in self.fields:
            self.groups = self.cache_model(model.CustomerGroup, 'id')

    def cache_query_options(self):
        options = []
        if 'first_name' in self.fields or 'last_name' in self.fields:
            options.append(orm.joinedload(model.Customer._person).joinedload(model.CustomerPerson.person))
        if 'phone_number' in self.fields or 'phone_number_2' in self.fields:
            options.append(orm.joinedload(model.Customer.phones))
        if 'email_address' in self.fields:
            options.append(orm.joinedload(model.Customer.email))
        if 'group_id' in self.fields or 'group_id_2' in self.fields:
            options.append(orm.joinedload_all(model.Customer._groups, model.CustomerGroupAssignment.group))
        return options

    def normalize_record(self, data):
        self.string_or_null(data, 'id', 'phone_number', 'phone_number_2',
                            'email_address', 'group_id', 'group_id_2')
        self.prioritize_2(data, 'phone_number')
        self.prioritize_2(data, 'group_id')

    def normalize_instance(self, customer):
        data = super(CustomerImporter, self).normalize_instance(customer)
        if 'first_name' in self.fields or 'last_name' in self.fields:
            person = customer.person
            if person:
                if 'first_name' in self.fields:
                    data['first_name'] = person.first_name
                if 'last_name' in self.fields:
                    data['last_name'] = person.last_name
            else:
                if 'first_name' in self.fields:
                    data['first_name'] = None
                if 'last_name' in self.fields:
                    data['last_name'] = None

        if 'phone_number' in self.fields or 'phone_number_2' in self.fields:
            phones = filter(lambda p: p.type == 'Voice', customer.phones)
        if 'phone_number' in self.fields:
            data['phone_number'] = phones[0].number if phones else None
        if 'phone_number_2' in self.fields:
            data['phone_number_2'] = phones[1].number if len(phones) > 1 else None

        if 'email_address' in self.fields:
            email = customer.email
            data['email_address'] = email.address if email else None

        if 'group_id' in self.fields:
            group = customer.groups[0] if customer.groups else None
            data['group_id'] = group.id if group else None

        if 'group_id_2' in self.fields:
            group = customer.groups[1] if customer.groups and len(customer.groups) > 1 else None
            data['group_id_2'] = group.id if group else None

        return data

    def update_instance(self, customer, data, inst_data=None):
        super(CustomerImporter, self).update_instance(customer, data, inst_data)
        if 'first_name' in data or 'last_name' in data:
            person = customer.person
            if not person:
                person = model.Person()
                customer.people.append(person)
        if 'first_name' in data:
            person.first_name = data['first_name']
        if 'last_name' in data:
            person.last_name = data['last_name']

        if 'phone_number' in data:
            phones = filter(lambda p: p.type == 'Voice', customer.phones)
            number = data['phone_number']
            if number:
                if phones:
                    phone = phones[0]
                    if phone.number != number:
                        phone.number = number
                else:
                    customer.add_phone_number(number, type='Voice')
            else:
                for phone in phones:
                    customer.phones.remove(phone)

        if 'phone_number_2' in data:
            phones = filter(lambda p: p.type == 'Voice', customer.phones)
            number = data['phone_number_2']
            if number:
                if len(phones) > 1:
                    phone = phones[1]
                    if phone.number != number:
                        phone.number = number
                else:
                    customer.add_phone_number(number, 'Voice')
            else:
                for phone in phones[1:]:
                    customer.phones.remove(phone)

        if 'email_address' in data:
            address = data['email_address']
            if address:
                if customer.email:
                    if customer.email.address != address:
                        customer.email.address = address
                else:
                    customer.add_email_address(address)
            else:
                if len(customer.emails):
                    del customer.emails[:]

        if 'group_id' in data:
            id_ = data['group_id']
            if id_:
                group = self.groups.get(id_)
                if not group:
                    group = model.CustomerGroup()
                    group.id = id_
                    group.name = "(auto-created)"
                    self.session.add(group)
                    self.groups[group.id] = group
                if group in customer.groups:
                    if group is not customer.groups[0]:
                        customer.groups.remove(group)
                        customer.groups.insert(0, group)
                else:
                    customer.groups.insert(0, group)
            else:
                if customer.groups:
                    del customer.groups[:]

        if 'group_id_2' in data:
            id_ = data['group_id_2']
            if id_:
                group = self.groups.get(id_)
                if not group:
                    group = model.CustomerGroup()
                    group.id_ = id_
                    group.name = "(auto-created)"
                    self.session.add(group)
                    self.groups[group.id] = group
                if group in customer.groups:
                    if len(customer.groups) > 1:
                        if group is not customer.groups[1]:
                            customer.groups.remove(group)
                            customer.groups.insert(1, group)
                else:
                    if len(customer.groups) > 1:
                        customer.groups.insert(1, group)
                    else:
                        customer.groups.append(group)
            else:
                if len(customer.groups) > 1:
                    del customer.groups[1:]


class CustomerPersonImporter(Importer):
    """
    CustomerPerson data importer.
    """
    model_class = model.CustomerPerson
    supported_fields = [
        'uuid',
        'customer_uuid',
        'person_uuid',
        'ordinal',
        ]


class CustomerPhoneNumberImporter(Importer):
    """
    Customer phone number data importer.
    """
    model_class = model.CustomerPhoneNumber
    supported_fields = [
        'uuid',
        'parent_uuid',
        'type',
        'preference',
        'number',
        ]


class VendorImporter(Importer):
    """
    Vendor data importer.
    """
    model_class = model.Vendor
    simple_fields = [
        'uuid',
        'id',
        'name',
        'special_discount',
        'lead_time_days',
        'order_interval_days',
        ]
    phone_fields = [
        'phone_number',
        'phone_number_2',
        'fax_number',
        'fax_number_2',
        ]
    contact_fields = [
        'contact_name',
        'contact_name_2',
        ]
    complex_fields = [
        'email_address',
        ]

    @property
    def supported_fields(self):
        return (self.simple_fields + self.phone_fields + self.contact_fields
                + self.complex_fields)

    def cache_query_options(self):
        options = []
        for field in self.phone_fields:
            if field in self.fields:
                options.append(orm.joinedload(model.Vendor.phones))
                break
        for field in self.contact_fields:
            if field in self.fields:
                options.append(orm.joinedload(model.Vendor._contacts))
                break
        if 'email_address' in self.fields:
            options.append(orm.joinedload(model.Vendor.email))
        return options

    def normalize_record(self, data):
        self.string_or_null(data, 'id', 'name', 'email_address',
                            'phone_number', 'phone_number_2',
                            'fax_number', 'fax_number_2',
                            'contact_name', 'contact_name_2')
        self.prioritize_2(data, 'phone_number')
        self.prioritize_2(data, 'fax_number')
        self.prioritize_2(data, 'contact_name')

    def normalize_instance(self, vendor):
        data = super(VendorImporter, self).normalize_instance(vendor)

        if 'phone_number' in self.fields:
            data['phone_number'] = None
            for phone in vendor.phones:
                if phone.type == 'Voice':
                    data['phone_number'] = phone.number
                    break

        if 'phone_number_2' in self.fields:
            data['phone_number_2'] = None
            first = False
            for phone in vendor.phones:
                if phone.type == 'Voice':
                    if first:
                        data['phone_number_2'] = phone.number
                        break
                    first = True

        if 'fax_number' in self.fields:
            data['fax_number'] = None
            for phone in vendor.phones:
                if phone.type == 'Fax':
                    data['fax_number'] = phone.number
                    break

        if 'fax_number_2' in self.fields:
            data['fax_number_2'] = None
            first = False
            for phone in vendor.phones:
                if phone.type == 'Fax':
                    if first:
                        data['fax_number_2'] = phone.number
                        break
                    first = True

        if 'contact_name' in self.fields:
            contact = vendor.contact
            data['contact_name'] = contact.display_name if contact else None

        if 'contact_name_2' in self.fields:
            contact = vendor.contacts[1] if len(vendor.contacts) > 1 else None
            data['contact_name_2'] = contact.display_name if contact else None

        if 'email_address' in self.fields:
            email = vendor.email
            data['email_address'] = email.address if email else None

        return data

    def update_instance(self, vendor, data, inst_data=None):
        super(VendorImporter, self).update_instance(vendor, data, inst_data)

        if 'phone_number' in data:
            number = data['phone_number'] or None
            if number:
                found = False
                for phone in vendor.phones:
                    if phone.type == 'Voice':
                        if phone.number != number:
                            phone.number = number
                        found = True
                        break
                if not found:
                    vendor.add_phone_number(number, type='Voice')
            else:
                for phone in list(vendor.phones):
                    if phone.type == 'Voice':
                        vendor.phones.remove(phone)

        if 'phone_number_2' in data:
            number = data['phone_number_2'] or None
            if number:
                found = False
                first = False
                for phone in vendor.phones:
                    if phone.type == 'Voice':
                        if first:
                            if phone.number != number:
                                phone.number = number
                            found = True
                            break
                        first = True
                if not found:
                    vendor.add_phone_number(number, type='Voice')
            else:
                first = False
                for phone in list(vendor.phones):
                    if phone.type == 'Voice':
                        if first:
                            vendor.phones.remove(phone)
                            break
                        first = True

        if 'fax_number' in data:
            number = data['fax_number'] or None
            if number:
                found = False
                for phone in vendor.phones:
                    if phone.type == 'Fax':
                        if phone.number != number:
                            phone.number = number
                        found = True
                        break
                if not found:
                    vendor.add_phone_number(number, type='Fax')
            else:
                for phone in list(vendor.phones):
                    if phone.type == 'Fax':
                        vendor.phones.remove(phone)

        if 'fax_number_2' in data:
            number = data['fax_number_2'] or None
            if number:
                found = False
                first = False
                for phone in vendor.phones:
                    if phone.type == 'Fax':
                        if first:
                            if phone.number != number:
                                phone.number = number
                            found = True
                            break
                        first = True
                if not found:
                    vendor.add_phone_number(number, type='Fax')
            else:
                first = False
                for phone in list(vendor.phones):
                    if phone.type == 'Fax':
                        if first:
                            vendor.phones.remove(phone)
                            break
                        first = True

        if 'contact_name' in data:
            if data['contact_name']:
                contact = vendor.contact
                if not contact:
                    contact = model.Person()
                    self.session.add(contact)
                    vendor.contacts.append(contact)
                contact.display_name = data['contact_name']
            else:
                if len(vendor.contacts):
                    del vendor.contacts[:]

        if 'contact_name_2' in data:
            if data['contact_name_2']:
                contact = vendor.contacts[1] if len(vendor.contacts) > 1 else None
                if not contact:
                    contact = model.Person()
                    self.session.add(contact)
                    vendor.contacts.append(contact)
                contact.display_name = data['contact_name_2']
            else:
                if len(vendor.contacts) > 1:
                    del vendor.contacts[1:]

        if 'email_address' in data:
            address = data['email_address'] or None
            if address:
                if vendor.email:
                    if vendor.email.address != address:
                        vendor.email.address = address
                else:
                    vendor.add_email_address(address)
            else:
                if len(vendor.emails):
                    del vendor.emails[:]


class DepartmentImporter(Importer):
    """
    Department data importer.
    """
    model_class = model.Department
    normalizer_class = normal.DepartmentNormalizer
    simple_fields = [
        'uuid',
        'number',
        'name',
    ]
    supported_fields = simple_fields


class SubdepartmentImporter(Importer):
    """
    Subdepartment data importer.
    """
    model_class = model.Subdepartment
    simple_fields = [
        'uuid',
        'number',
        'name',
        'department_uuid',
        ]
    supported_fields = simple_fields + [
        'department_number',
        ]

    def setup(self):
        self.departments = self.cache_model(model.Department, 'number')

    def cache_query_options(self):
        if 'department_number' in self.fields:
            return [orm.joinedload(model.Subdepartment.department)]

    def normalize_record(self, data):
        self.string_or_null(data, 'name')

    def normalize_instance(self, subdepartment):
        data = super(SubdepartmentImporter, self).normalize_instance(subdepartment)
        if 'department_number' in self.fields:
            dept = subdepartment.department
            data['department_number'] = dept.number if dept else None
        return data

    def update_instance(self, subdepartment, data, inst_data=None):
        super(SubdepartmentImporter, self).update_instance(subdepartment, data, inst_data)
        if 'department_number' in self.fields:
            dept = self.departments.get(data['department_number'])
            if not dept:
                dept = model.Department()
                dept.number = data['department_number']
                dept.name = "(auto-created)"
                self.session.add(dept)
                self.departments[dept.number] = dept
            subdepartment.department = dept


class CategoryImporter(Importer):
    """
    Category data importer.
    """
    model_class = model.Category
    supported_fields = [
        'uuid',
        'number',
        'name',
        'department_uuid',
        ]

    def normalize_record(self, data):
        self.string_or_null(data, 'name')


class FamilyImporter(Importer):
    """
    Family data importer.
    """
    model_class = model.Family
    supported_fields = [
        'uuid',
        'code',
        'name',
        ]

    def normalize_record(self, data):
        self.string_or_null(data, 'name')


class ReportCodeImporter(Importer):
    """
    ReportCode data importer.
    """
    model_class = model.ReportCode
    supported_fields = [
        'uuid',
        'code',
        'name',
        ]

    def normalize_record(self, data):
        self.int_or_null(data, 'code')
        self.string_or_null(data, 'name')


class DepositLinkImporter(Importer):
    """
    Deposit link data importer.
    """
    model_class = model.DepositLink
    supported_fields = [
        'uuid',
        'code',
        'description',
        'amount',
        ]

    def normalize_record(self, data):
        self.int_or_null(data, 'code')
        self.string_or_null(data, 'description')


class TaxImporter(Importer):
    """
    Tax data importer.
    """
    model_class = model.Tax
    supported_fields = [
        'uuid',
        'code',
        'description',
        'rate',
        ]

    def normalize_record(self, data):
        self.int_or_null(data, 'code')
        self.string_or_null(data, 'description')


class BrandImporter(Importer):
    """
    Brand data importer.
    """
    model_class = model.Brand
    supported_fields = [
        'uuid',
        'name',
        ]

    def normalize_record(self, data):
        self.string_or_null(data, 'name')


class ProductImporter(Importer):
    """
    Data importer for :class:`rattail.db.model.Product`.
    """
    model_class = model.Product
    simple_fields = [
        'uuid',
        'upc',
        'description',
        'unit_size',
        'size',
        'department_uuid',
        'subdepartment_uuid',
        'category_uuid',
        'family_uuid',
        'report_code_uuid',
        'deposit_link_uuid',
        'tax_uuid',
        'brand_uuid',
        'unit_of_measure',
        'case_pack',
        'weighed',
        'organic',
        'discountable',
        'special_order',
        'not_for_sale',
        'last_sold',
        ]
    regular_price_fields = [
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_pack_price',
        'regular_price_pack_multiple',
        'regular_price_type',
        'regular_price_level',
        'regular_price_starts',
        'regular_price_ends',
        ]
    sale_price_fields = [
        'sale_price_price',
        'sale_price_multiple',
        'sale_price_pack_price',
        'sale_price_pack_multiple',
        'sale_price_type',
        'sale_price_level',
        'sale_price_starts',
        'sale_price_ends',
        ]
    supported_fields = simple_fields + regular_price_fields + sale_price_fields + [
        'brand_name',
        'department_number',
        'subdepartment_number',
        'category_number',
        'family_code',
        'report_code',
        'deposit_link_code',
        'tax_code',
        'vendor_id',
        'vendor_item_code',
        'vendor_case_cost',
        ]

    def setup(self):
        if 'brand_name' in self.fields:
            self.brands = self.cache_model(model.Brand, 'name')
        if 'department_number' in self.fields:
            self.departments = self.cache_model(model.Department, 'number')
        if 'subdepartment_number' in self.fields:
            self.subdepartments = self.cache_model(model.Subdepartment, 'number')
        if 'category_number' in self.fields:
            self.categories = self.cache_model(model.Category, 'number')
        if 'family_code' in self.fields:
            self.families = self.cache_model(model.Family, 'code')
        if 'report_code' in self.fields:
            self.reportcodes = self.cache_model(model.ReportCode, 'code')
        if 'deposit_link_code' in self.fields:
            self.depositlinks = self.cache_model(model.DepositLink, 'code')
        if 'tax_code' in self.fields:
            self.taxes = self.cache_model(model.Tax, 'code')
        if 'vendor_id' in self.fields:
            self.vendors = self.cache_model(model.Vendor, 'id')

    def cache_query_options(self):
        options = []
        if 'brand_name' in self.fields:
            options.append(orm.joinedload(model.Product.brand))
        if 'department_number' in self.fields:
            options.append(orm.joinedload(model.Product.department))
        if 'subdepartment_number' in self.fields:
            options.append(orm.joinedload(model.Product.subdepartment))
        if 'category_number' in self.fields:
            options.append(orm.joinedload(model.Product.category))
        if 'family_code' in self.fields:
            options.append(orm.joinedload(model.Product.family))
        if 'report_code' in self.fields:
            options.append(orm.joinedload(model.Product.report_code))
        if 'deposit_link_code' in self.fields:
            options.append(orm.joinedload(model.Product.deposit_link))
        if 'tax_code' in self.fields:
            options.append(orm.joinedload(model.Product.tax))
        joined_prices = False
        for field in self.regular_price_fields:
            if field in self.fields:
                options.append(orm.joinedload(model.Product.prices))
                options.append(orm.joinedload(model.Product.regular_price))
                joined_prices = True
                break
        for field in self.sale_price_fields:
            if field in self.fields:
                if not joined_prices:
                    options.append(orm.joinedload(model.Product.prices))
                options.append(orm.joinedload(model.Product.current_price))
                break
        if ('vendor_id' in self.fields
            or 'vendor_item_code' in self.fields
            or 'vendor_case_cost' in self.fields):
            options.append(orm.joinedload(model.Product.cost))
        return options

    def normalize_record(self, data):
        self.string_or_null(data, 'brand_name', 'description', 'unit_of_measure',
                            'vendor_id', 'vendor_item_code')
        self.int_or_null(data, 'family_code', 'report_code', 'deposit_link_code', 'tax_code')

    def normalize_instance(self, product):
        data = super(ProductImporter, self).normalize_instance(product)
        if 'brand_name' in self.fields:
            data['brand_name'] = product.brand.name if product.brand else None
        if 'department_number' in self.fields:
            data['department_number'] = product.department.number if product.department else None
        if 'subdepartment_number' in self.fields:
            data['subdepartment_number'] = product.subdepartment.number if product.subdepartment else None
        if 'category_number' in self.fields:
            data['category_number'] = product.category.number if product.category else None
        if 'family_code' in self.fields:
            data['family_code'] = product.family.code if product.family else None
        if 'report_code' in self.fields:
            data['report_code'] = product.report_code.code if product.report_code else None
        if 'deposit_link_code' in self.fields:
            data['deposit_link_code'] = product.deposit_link.code if product.deposit_link else None
        if 'tax_code' in self.fields:
            data['tax_code'] = product.tax.code if product.tax else None

        for field in self.regular_price_fields:
            if field in self.fields:
                price = product.regular_price
                if 'regular_price_price' in self.fields:
                    data['regular_price_price'] = price.price if price else None
                if 'regular_price_multiple' in self.fields:
                    data['regular_price_multiple'] = price.multiple if price else None
                if 'regular_price_pack_price' in self.fields:
                    data['regular_price_pack_price'] = price.pack_price if price else None
                if 'regular_price_pack_multiple' in self.fields:
                    data['regular_price_pack_multiple'] = price.pack_multiple if price else None
                if 'regular_price_type' in self.fields:
                    data['regular_price_type'] = price.type if price else None
                if 'regular_price_level' in self.fields:
                    data['regular_price_level'] = price.level if price else None
                if 'regular_price_starts' in self.fields:
                    data['regular_price_starts'] = price.starts if price else None
                if 'regular_price_ends' in self.fields:
                    data['regular_price_ends'] = price.ends if price else None
                break

        for field in self.sale_price_fields:
            if field in self.fields:
                price = product.current_price
                if 'sale_price_price' in self.fields:
                    data['sale_price_price'] = price.price if price else None
                if 'sale_price_multiple' in self.fields:
                    data['sale_price_multiple'] = price.multiple if price else None
                if 'sale_price_pack_price' in self.fields:
                    data['sale_price_pack_price'] = price.pack_price if price else None
                if 'sale_price_pack_multiple' in self.fields:
                    data['sale_price_pack_multiple'] = price.pack_multiple if price else None
                if 'sale_price_type' in self.fields:
                    data['sale_price_type'] = price.type if price else None
                if 'sale_price_level' in self.fields:
                    data['sale_price_level'] = price.level if price else None
                if 'sale_price_starts' in self.fields:
                    data['sale_price_starts'] = price.starts if price else None
                if 'sale_price_ends' in self.fields:
                    data['sale_price_ends'] = price.ends if price else None
                break

        if 'vendor_id' in self.fields or 'vendor_item_code' in self.fields or 'vendor_case_cost' in self.fields:
            cost = product.cost
            if 'vendor_id' in self.fields:
                data['vendor_id'] = cost.vendor.id if cost else None
            if 'vendor_item_code' in self.fields:
                data['vendor_item_code'] = cost.code if cost else None
            if 'vendor_case_cost' in self.fields:
                data['vendor_case_cost'] = cost.case_cost if cost else None

        return data

    def update_instance(self, product, data, inst_data=None):
        super(ProductImporter, self).update_instance(product, data, inst_data)

        if 'brand_name' in data:
            name = data['brand_name']
            if name:
                brand = self.brands.get(name)
                if not brand:
                    brand = model.Brand()
                    brand.name = name
                    self.session.add(brand)
                    self.brands[brand.name] = brand
                product.brand = brand
            else:
                if product.brand:
                    product.brand = None

        if 'department_number' in data:
            number = data['department_number']
            if number:
                dept = self.departments.get(number)
                if not dept:
                    dept = model.Department()
                    dept.number = number
                    dept.name = "(auto-created)"
                    self.session.add(dept)
                    self.departments[dept.number] = dept
                product.department = dept
            else:
                if product.department:
                    product.department = None

        if 'subdepartment_number' in data:
            number = data['subdepartment_number']
            if number:
                sub = self.subdepartments.get(number)
                if not sub:
                    sub = model.Subdepartment()
                    sub.number = number
                    sub.name = "(auto-created)"
                    self.session.add(sub)
                    self.subdepartments[number] = sub
                product.subdepartment = sub
            else:
                if product.subdepartment:
                    product.subdepartment = None

        if 'category_number' in data:
            number = data['category_number']
            if number:
                cat = self.categories.get(number)
                if not cat:
                    cat = model.Category()
                    cat.number = number
                    cat.name = "(auto-created)"
                    self.session.add(cat)
                    self.categories[number] = cat
                product.category = cat
            else:
                if product.category:
                    product.category = None

        if 'family_code' in data:
            code = data['family_code']
            if code:
                family = self.families.get(code)
                if not family:
                    family = model.Family()
                    family.code = code
                    family.name = "(auto-created)"
                    self.session.add(family)
                    self.families[family.code] = family
                product.family = family
            else:
                if product.family:
                    product.family = None

        if 'report_code' in data:
            code = data['report_code']
            if code:
                rc = self.reportcodes.get(code)
                if not rc:
                    rc = model.ReportCode()
                    rc.code = code
                    rc.name = "(auto-created)"
                    self.session.add(rc)
                    self.reportcodes[rc.code] = rc
                product.report_code = rc
            else:
                if product.report_code:
                    product.report_code = None

        if 'deposit_link_code' in data:
            code = data['deposit_link_code']
            if code:
                link = self.depositlinks.get(code)
                if not link:
                    link = model.DepositLink()
                    link.code = code
                    link.description = "(auto-created)"
                    self.session.add(link)
                    self.depositlinks[link.code] = link
                product.deposit_link = link
            else:
                if product.deposit_link:
                    product.deposit_link = None

        if 'tax_code' in data:
            code = data['tax_code']
            if code:
                tax = self.taxes.get(code)
                if not tax:
                    tax = model.Tax()
                    tax.code = code
                    tax.description = "(auto-created)"
                    tax.rate = 0
                    self.session.add(tax)
                    self.taxes[tax.code] = tax
                product.tax = tax
            elif product.tax:
                product.tax = None

        create = False
        delete = False
        for field in self.regular_price_fields:
            if field in data:
                delete = True
                if data[field] is not None:
                    create = True
                    break
        if create:
            price = product.regular_price
            if not price:
                price = model.ProductPrice()
                product.prices.append(price)
                product.regular_price = price
            if 'regular_price_price' in data:
                price.price = data['regular_price_price']
            if 'regular_price_multiple' in data:
                price.multiple = data['regular_price_multiple']
            if 'regular_price_pack_price' in data:
                price.pack_price = data['regular_price_pack_price']
            if 'regular_price_pack_multiple' in data:
                price.pack_multiple = data['regular_price_pack_multiple']
            if 'regular_price_type' in data:
                price.type = data['regular_price_type']
            if 'regular_price_level' in data:
                price.level = data['regular_price_level']
            if 'regular_price_starts' in data:
                price.starts = data['regular_price_starts']
            if 'regular_price_ends' in data:
                price.ends = data['regular_price_ends']
        elif delete:
            if product.regular_price:
                product.regular_price = None

        create = False
        delete = False
        for field in self.sale_price_fields:
            if field in data:
                delete = True
                if data[field]:
                    create = True
                    break
        if create:
            price = product.current_price
            if not price:
                price = model.ProductPrice()
                product.prices.append(price)
                product.current_price = price
            if 'sale_price_price' in data:
                price.price = data['sale_price_price']
            if 'sale_price_multiple' in data:
                price.multiple = data['sale_price_multiple']
            if 'sale_price_pack_price' in data:
                price.pack_price = data['sale_price_pack_price']
            if 'sale_price_pack_multiple' in data:
                price.pack_multiple = data['sale_price_pack_multiple']
            if 'sale_price_type' in data:
                price.type = data['sale_price_type']
            if 'sale_price_level' in data:
                price.level = data['sale_price_level']
            if 'sale_price_starts' in data:
                price.starts = data['sale_price_starts']
            if 'sale_price_ends' in data:
                price.ends = data['sale_price_ends']
        elif delete:
            if product.current_price:
                product.current_price = None

        if 'vendor_id' in data:
            id_ = data['vendor_id']
            if id_:
                vendor = self.vendors.get(id_)
                if not vendor:
                    vendor = model.Vendor()
                    vendor.id = id_
                    vendor.name = "(auto-created)"
                    self.session.add(vendor)
                    self.vendors[id_] = vendor
                if product.cost:
                    if product.cost.vendor is not vendor:
                        cost = product.cost_for_vendor(vendor)
                        if not cost:
                            cost = model.ProductCost()
                            cost.vendor = vendor
                        product.costs.insert(0, cost)
                else:
                    cost = model.ProductCost()
                    cost.vendor = vendor
                    product.costs.append(cost)
                    # TODO: This seems heavy-handed, but also seems necessary
                    # to populate the `Product.cost` relationship...
                    self.session.add(product)
                    self.session.flush()
                    self.session.refresh(product)
            else:
                if product.cost:
                    del product.costs[:]

        if 'vendor_item_code' in self.fields:
            code = data['vendor_item_code']
            if data.get('vendor_id'):
                if product.cost:
                    product.cost.code = code
                else:
                    log.warning("product has no cost, so can't set vendor_item_code: {0}".format(product))

        if 'vendor_case_cost' in self.fields:
            cost = data['vendor_case_cost']
            if data.get('vendor_id'):
                if product.cost:
                    product.cost.case_cost = cost
                else:
                    log.warning("product has no cost, so can't set vendor_case_cost: {0}".format(product))


class ProductCodeImporter(Importer):
    """
    Data importer for :class:`rattail.db.model.ProductCode`.
    """
    model_class = model.ProductCode
    simple_fields = [
        'uuid',
        'product_uuid',
        'ordinal',
        'code',
        ]
    supported_fields = simple_fields + [
        'product_upc',
        'primary',
        ]

    def setup(self):
        if 'product_upc' in self.fields:
            self.products = self.cache_model(model.Product, 'upc')

    def cache_query_options(self):
        if 'product_upc' in self.fields:
            return [orm.joinedload(model.ProductCode.product)]

    def normalize_record(self, data):
        self.string_or_null(data, 'code')

    def normalize_instance(self, code):
        data = super(ProductCodeImporter, self).normalize_instance(code)
        if 'product_upc' in self.fields:
            data['product_upc'] = code.product.upc
        if 'primary' in self.fields:
            data['primary'] = code.ordinal == 1
        return data

    def update_instance(self, code, data, inst_data=None):
        super(ProductCodeImporter, self).update_instance(code, data, inst_data)

        if 'product_upc' in data and 'product_uuid' not in data:
            upc = data['product_upc']
            assert upc, "Source data has no product_upc value: {0}".format(repr(data))
            product = self.products.get(upc)
            if not product:
                product = model.Product()
                product.upc = upc
                product.description = "(auto-created)"
                self.session.add(product)
                self.products[product.upc] = product
                product._codes.append(code)
            else:
                if code not in product._codes:
                    product._codes.append(code)

        if 'primary' in data:
            if data['primary']:
                if code.ordinal != 1:
                    product = code.product
                    product._codes.remove(code)
                    product._codes.insert(0, code)
                    product._codes.reorder()
            elif code.ordinal == 1:
                product = code.product
                if len(product._codes) > 1:
                    product._codes.remove(code)
                    product._codes.append(code)
                    product._codes.reorder()


class ProductCostImporter(Importer):
    """
    Data importer for :class:`rattail.db.model.ProductCost`.
    """
    model_class = model.ProductCost
    simple_fields = [
        'uuid',
        'product_uuid',
        'vendor_uuid',
        'preference',
        'code',
        'case_size',
        'case_cost',
        'pack_size',
        'pack_cost',
        'unit_cost',
        'effective',
        ]
    supported_fields = simple_fields + [
        'product_upc',
        'vendor_id',
        'preferred',
        ]

    def setup(self):
        if 'product_upc' in self.fields:
            self.products = self.cache_model(model.Product, 'upc')
        if 'vendor_id' in self.fields:
            self.vendors = self.cache_model(model.Vendor, 'id')

    def cache_query_options(self):
        options = []
        if 'product_upc' in self.fields:
            options.append(orm.joinedload(model.ProductCost.product))
        if 'vendor_id' in self.fields:
            options.append(orm.joinedload(model.ProductCost.vendor))
        return options

    def normalize_instance(self, cost):
        data = super(ProductCostImporter, self).normalize_instance(cost)
        if 'product_upc' in self.fields:
            data['product_upc'] = cost.product.upc
        if 'vendor_id' in self.fields:
            data['vendor_id'] = cost.vendor.id
        data['preferred'] = cost.preference == 1
        return data
        
    def update_instance(self, cost, data, inst_data=None):
        super(ProductCostImporter, self).update_instance(cost, data, inst_data)

        if 'product_upc' in data and 'product_uuid' not in data:
            upc = data['product_upc']
            assert upc, "Source data has no product_upc value: {0}".format(repr(data))
            product = self.products.get(upc)
            if not product:
                product = model.Product()
                product.upc = upc
                product.description = "(auto-created)"
                self.session.add(product)
                self.products[product.upc] = product
            if not cost.product:
                product.costs.append(cost)
            elif cost.product is not product:
                log.warning("duplicate products detected for UPC {0}".format(upc.pretty()))

        if 'vendor_id' in data and 'vendor_uuid' not in data:
            id_ = data['vendor_id']
            assert id_, "Source data has no vendor_id value: {0}".format(repr(data))
            vendor = self.vendors.get(id_)
            if not vendor:
                vendor = model.Vendor()
                vendor.id = id_
                vendor.name = "(auto-created)"
                self.session.add(vendor)
                self.vendors[vendor.id] = vendor
            cost.vendor = vendor

        if 'preferred' in data:
            if data['preferred']:
                if cost.preference != 1:
                    product = cost.product
                    product.costs.remove(cost)
                    product.costs.insert(0, cost)
            else:
                if cost.preference == 1:
                    product = cost.product
                    if len(product.costs) > 1:
                        product.costs.remove(cost)
                        product.costs.append(cost)
                        product.costs.reorder()



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Consumers
"""

from __future__ import unicode_literals, absolute_import

from rattail.config import parse_list
from rattail.db.newimporting import ImportHandler


class DataSyncConsumer(object):
    """
    Base class for all DataSync consumers.
    """

    def __init__(self, config, key, dbkey=None):
        self.config = config
        self.key = key
        self.dbkey = dbkey

    def setup(self):
        """
        This method is called when the consumer thread is first started.
        """

    def begin_transaction(self):
        """
        Called just before the consumer is asked to process changes, possibly
        via multiple batches.
        """

    def process_changes(self, session, changes):
        """
        Process (consume) a batch of changes.
        """

    def rollback_transaction(self):
        """
        Called when any batch of changes failed to process.
        """

    def commit_transaction(self):
        """
        Called just after the consumer has successfully finished processing
        changes, possibly via multiple batches.
        """


class DataSyncImportConsumer(DataSyncConsumer):
    """
    Base class for DataSync consumer which is able to leverage a (set of)
    importer(s) to do the heavy lifting.
    """

    def __init__(self, *args, **kwargs):
        super(DataSyncImportConsumer, self).__init__(*args, **kwargs)
        self.importers = self.get_importers()

    def get_importers(self):
        """
        You must override this to return a dict of importer *instances*, keyed
        by what you expect the corresponding ``DataSyncChange.payload_type`` to
        be, coming from the "host" system, whatever that is.
        """
        raise NotImplementedError

    def get_importers_from_handler(self, handler, default_only=True):
        if not isinstance(handler, ImportHandler):
            handler = handler(config=self.config)
        factories = handler.get_importers()
        if default_only:
            keys = handler.get_default_keys()
        else:
            keys = factories.keys()
        importers = {}
        for key in keys:
            importers[key] = factories[key](config=self.config)
        return importers

    def process_changes(self, session, changes):
        """
        Process all changes, leveraging importer(s) as much as possible.
        """
        # Update all importers with current Rattail session.
        for importer in self.importers.itervalues():
            importer.session = session

        for change in changes:
            self.invoke_importer(session, change)

    def invoke_importer(self, session, change):
        """
        For the given change, invoke the default importer behavior, if one is
        available.
        """
        importer = self.importers.get(change.payload_type)
        if importer:
            if change.deletion:
                self.process_deletion(session, importer, change)
            else:
                return self.process_change(session, importer, change)

    def process_change(self, session, importer, change=None, host_object=None, host_data=None):
        """
        Invoke the importer to process the given change / host record.
        """
        if host_data is None:
            if host_object is None:
                host_object = self.get_host_record(session, change)
                if host_object is None:
                    return
            host_data = importer.normalize_source_record(host_object)
            if host_data is None:
                return
        key = importer.get_key(host_data)
        local_object = importer.get_instance(key)
        if local_object:
            local_data = importer.normalize_instance(local_object)
            if importer.data_diffs(local_data, host_data):
                local_object = importer.update_instance(local_object, host_data, local_data)
            return local_object
        else:
            return importer.create_instance(key, host_data)

    def process_deletion(self, session, importer, change):
        """
        Attempt to invoke the importer, to delete a local record according to
        the change involved.
        """
        key = self.get_deletion_key(session, change)
        local_object = importer.get_instance(key)
        if local_object:
            return importer.delete_instance(local_object)
        return False

    def get_deletion_key(self, session, change):
        return (change.payload_key,)

    def get_host_record(self, session, change):
        """
        You must override this, to return a host record from the given
        ``DataSyncChange`` instance.  Note that the host record need *not* be
        normalized, as that will be done by the importer.  (This is effectively
        the only part of the processing which is not handled by the importer.)
        """
        raise NotImplementedError


class ErrorTestConsumer(DataSyncConsumer):
    """
    Consumer which always raises an error when processing changes.  Useful for
    testing error handling etc.
    """

    def process_changes(self, session, changes):
        raise RuntimeError("Fake exception, to test error handling")



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail->Rattail Data Import
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import Session
from rattail.db.importing import ImportHandler, QueryProvider, models, normal
from rattail.util import OrderedDict


class RattailImportHandler(ImportHandler):
    """
    Import handler for data coming from a Rattail database.
    """

    def get_importers(self):
        importers = OrderedDict()
        importers['Message'] = MessageProvider
        importers['MessageRecipient'] = MessageRecipientProvider
        return importers

    def before_import(self):
        self.source_session = Session()

    def get_importer_kwargs(self, key):
        return {'source_session': self.source_session}

    def after_import(self):
        self.source_session.close()
        del self.source_session


class RattailProvider(QueryProvider):
    """
    Base class for Rattail data providers.
    """
    key = 'uuid'

    @property
    def normalize_progress_message(self):
        return "Reading {} data from Rattail".format(self.model_name)

    def setup(self):
        self.normalizer = self.normalizer_class()

    def query(self):
        return self.source_session.query(self.model_class)

    def normalize(self, instance):
        return self.normalizer.normalize(instance)


class MessageProvider(RattailProvider):
    """
    Import provider for Rattail message data.
    """
    importer_class = models.MessageImporter
    normalizer_class = normal.MessageNormalizer
    supported_fields = [
        'uuid',
        'sender_uuid',
        'subject',
        'body',
        'sent',
    ]


class MessageRecipientProvider(RattailProvider):
    """
    Import provider for Rattail message recipient data.
    """
    importer_class = models.MessageRecipientImporter
    normalizer_class = normal.MessageRecipientNormalizer
    supported_fields = [
        'uuid',
        'message_uuid',
        'recipient_uuid',
        'status',
    ]



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Vendor Invoice batches
"""

from __future__ import unicode_literals

import os

from sqlalchemy.orm import joinedload

from rattail.db import api
from rattail.db import model
from rattail.db.batch.handler import FileBatchHandler
from rattail.vendors.invoices import require_invoice_parser


class VendorInvoiceHandler(FileBatchHandler):
    """
    Handler for vendor invoice batches.
    """
    batch_model_class = model.VendorInvoice
    show_progress = True
    po_number_title = "PO Number"

    def make_batch(self, session, path, **kwargs):
        parser_key = kwargs.get('parser_key')
        if parser_key and not kwargs.get('vendor') and not kwargs.get('vendor_uuid'):
            parser = require_invoice_parser(parser_key)
            kwargs['vendor'] = api.get_vendor(session, parser.vendor_key)
        return super(VendorInvoiceHandler, self).make_batch(session, path, **kwargs)

    def get_purchase_order(self, number):
        """
        Fetch the purchase order object corresponding to the given PO number.
        Custom handlers should override this if aiming to reconcile invoices to
        purchase orders.
        """

    def validate_po_number(self, number):
        """
        This method does nothing by default.  Derived handlers can validate the
        PO number as they like.  Invalid PO numbers should cause ``ValueError``
        to be raised, with text of the reason for validation failure.
        """

    def refresh_data(self, session, batch, progress=None):
        """
        Refresh all data for the batch.
        """
        self.session = session
        del batch.data_rows[:]
        data_path = self.data_path(batch)
        parser = require_invoice_parser(batch.parser_key)
        parser.session = session
        parser.vendor = api.get_vendor(session, parser.vendor_key)
        batch.invoice_date = parser.parse_invoice_date(data_path)

        # Pre-cache products by UPC and vendor code.
        self.vendor = parser.vendor
        self.products = {'upc': {}, 'vendor_code': {}, 'code': {}}
        products = session.query(model.Product)\
            .options(joinedload(model.Product.brand))\
            .options(joinedload(model.Product.costs))\
            .options(joinedload(model.Product._codes))
        prog = None
        if progress:
            prog = progress("Caching products by UPC and vendor item code", products.count())
        for i, product in enumerate(products, 1):
            if product.upc:
                self.products['upc'][product.upc] = product
            cost = product.cost_for_vendor(self.vendor)
            product.vendor_cost = cost
            if cost and cost.code:
                self.products['vendor_code'][cost.code] = product
            for code in product.codes:
                self.products['code'][code] = product
            if prog:
                prog.update(i)
        if prog:
            prog.destroy()

        # Get data from parser, and convert to rows.
        data = list(parser.parse_rows(data_path))
        self.make_rows(session, batch, data, progress=progress)

        # If we have a PO number, attempt to reconcile against a purchase order.
        if batch.purchase_order_number:
            purchase = self.get_purchase_order(batch.purchase_order_number)
            if purchase:
                self.cognize_purchase_order(session, batch, purchase, progress=progress)

    def find_product(self, row):
        """
        Attempt to locate the product for the row, based on UPC etc.
        """
        if row.upc:
            product = self.products['upc'].get(row.upc)
            if product:
                return product

        if row.vendor_code:
            product = self.products['vendor_code'].get(row.vendor_code)
            if product:
                return product

    def cognize_row(self, session, row):
        """
        Inspect a single row from a invoice, and set its attributes based on
        whether or not the product exists, if we already have a cost record for
        the vendor, if the invoice contains a change etc.  Note that the
        product lookup is done first by UPC and then by vendor item code.
        """
        product = self.find_product(row)
        if not product:
            row.status_code = row.STATUS_NOT_IN_DB
            return

        row.product = product
        row.upc = product.upc
        row.brand_name = product.brand.name if product.brand else None
        row.description = product.description
        row.size = product.size
        if product.cost_for_vendor(self.vendor):
            row.status_code = row.STATUS_OK
        else:
            row.status_code = row.STATUS_COST_NOT_IN_DB

        # Calculate case cost if the parser couldn't provide one.
        if not row.case_cost and row.unit_cost and row.case_quantity:
            row.case_cost = row.unit_cost * row.case_quantity

    def cognize_purchase_order(self, session, invoice, purchase, progress=None):
        """
        Cognize the given invoice against the given purchase order object.
        Custom handlers should override this if aiming to reconcile invoices to
        purchase orders.
        """

    def execute(self, batch, progress=None):
        """
        Execute the vendor invoice batch.  Note that the default handler does
        not perform any actions; a custom handler must be used for anything
        interesting to happen.
        """
        return True



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase

from rattail import core


class TestCore(TestCase):

    def test_get_uuid(self):
        uuid = core.get_uuid()
        self.assertTrue(isinstance(uuid, str))
        self.assertEqual(len(uuid), 32)
